from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class SubsetOffset(QueryParameter):
    """The first object index to retrieve in `offset` mode.

    If the value of `subset` is not `"offset"`, this attribute will
    be ignored.
    """
    name = 'offset'
    schema = dict(type='integer', default=0)


register_global_parameter(SubsetOffset)



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class MediaRedirect(QueryParameter):
    """if `true`, make a redirect, otherwise return the file URL.
    """
    name = 'redirect'
    schema = dict(type='boolean', default=False)


register_global_parameter(MediaRedirect)



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class Subset(QueryParameter):
    """Subset mode for objects list.

    * if `null` (default value), you receive a list of objects
    without subsets management.
    * if `"pagination"`, you receive a list of objects
    with pagination management (see parameters:
    `page`, `page_size`, `page_subset`). The following attributes are add
    to `meta.subset` in response :
        * `page` : the page number
        * `page_size` : the max number of objets per pages
        * `count` : the total number of objects
        * `page_count` : the total number of pages
        * `subset` : a limited list of page numbers (for paginator)
        * `page_subset` : max number of pages in `subset`
    * if `"offset"`, you receive a list of objects. The following attributes
    are add to `meta.subset` in response :
        * `count` : the total number of objects
        * `page_count` : the total number of pages
        * `limit` : the max number of objets in result
        * `offset` : the first object index in result
    with offset management (see parameters: `page_size`)

    The `PFX_MAX_LIST_RESULT_SIZE` will always be applied is set to limit
    the result list size.
    """
    name = 'subset'
    schema = dict(
        type='string', default=None, enum=['pagination', 'offset', None])


register_global_parameter(Subset)



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class ListMode(QueryParameter):
    """Mode of result list.

    * If `"list"` (default value): Full object representation for lists.
    * If `"select"`: Minimal object representation for select widget.
    """
    name = 'mode'
    schema = dict(type='string', default='list', enum=['list', 'select'])


register_global_parameter(ListMode)



from django.db import connection, models
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import include, path
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore import register_views
from pfx.pfxcore.decorator import rest_view
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.test import APIClient, TestAssertMixin
from pfx.pfxcore.views import RestView
from tests.views import FakeViewMixin

# Add i18n string for Value 2, to check this is not used to translate
# the same value in choices.
_("Value 2")


class TestChoicesModel(JSONReprMixin, models.Model):
    choice1 = models.CharField(max_length=30, choices=[
        ('v1', "Value 1"),
        ('v2', "Value 2"),
    ])
    choice2 = models.CharField(max_length=30, choices=[
        ('v3', _("Value 3")),
        ('v4', _("Value 4")),
    ])

    class Meta:
        verbose_name = "TestModel"
        verbose_name_plural = "TestModels"
        ordering = ['pk']


@rest_view("/test")
class TestChoicesRestView(FakeViewMixin, RestView):
    default_public = True
    model = TestChoicesModel


urlpatterns = [
    path('api/', include(register_views(TestChoicesRestView))),
    path('api/', include('pfx.pfxcore.urls'))
]


@override_settings(ROOT_URLCONF=__name__)
class TestFieldsChoices(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(TestChoicesModel)

    def test_meta(self):
        response = self.client.get('/api/test/meta', locale='fr')
        self.assertRC(response, 200)
        self.assertJE(response, 'fields.choice1.choices.@0.label', "Value 1")
        self.assertJE(response, 'fields.choice1.choices.@1.label', "Value 2")
        self.assertJE(response, 'fields.choice2.choices.@0.label', "Valeur 3")
        self.assertJE(response, 'fields.choice2.choices.@1.label', "Valeur 4")

    def test_meta_list(self):
        response = self.client.get('/api/test/meta/list', locale='fr')
        self.assertRC(response, 200)
        self.assertJE(response, 'fields.choice1.choices.@0.label', "Value 1")
        self.assertJE(response, 'fields.choice1.choices.@1.label', "Value 2")
        self.assertJE(response, 'fields.choice2.choices.@0.label', "Valeur 3")
        self.assertJE(response, 'fields.choice2.choices.@1.label', "Valeur 4")

    def test_get(self):
        c1 = TestChoicesModel.objects.create(choice1='v2', choice2='v3')

        response = self.client.get(f'/api/test/{c1.pk}', locale='fr')
        self.assertRC(response, 200)
        self.assertJE(response, 'choice1.label', "Value 2")
        self.assertJE(response, 'choice2.label', "Valeur 3")

    def test_get_list(self):
        TestChoicesModel.objects.create(choice1='v2', choice2='v3')

        response = self.client.get('/api/test', locale='fr')
        self.assertRC(response, 200)
        self.assertJE(response, 'items.@0.choice1.label', "Value 2")
        self.assertJE(response, 'items.@0.choice2.label', "Valeur 3")



from django.db import models


class NotNullFieldMixin:
    def __init__(self, *args, **kwargs):
        kwargs['null'] = False
        super().__init__(*args, **kwargs)

    def get_prep_value(self, value):
        return "" if value is None else value


class NotNullCharField(NotNullFieldMixin, models.CharField):
    pass


class NotNullTextField(NotNullFieldMixin, models.TextField):
    pass


class NotNullURLField(NotNullFieldMixin, models.URLField):
    pass



import cProfile
import io
import logging
import pstats
from contextvars import ContextVar
from time import monotonic_ns

from django.db import connection, connections
from django.utils.connection import BaseConnectionHandler
from django.utils.deprecation import MiddlewareMixin

import sqlparse

from pfx.pfxcore.settings import settings

logger = logging.getLogger(__name__)
rid = ContextVar('var')


class ProfilingDatabaseRouter:
    def db_for_read(self, model, **hints):
        return rid.get('default')

    def db_for_write(self, model, **hints):
        return rid.get('default')


class ProfilingMiddleware(MiddlewareMixin):
    def process_request(self, request):
        request._profile = None
        if settings.PFX_PROFILING_CPROFILE and request.path.startswith(
                settings.PFX_PROFILING_CPROFILE):
            request._profile = cProfile.Profile()
            request._profile.enable()

        request._db_router = False
        if 'ProfilingDatabaseRouter' in ','.join(settings.DATABASE_ROUTERS):
            request._db_router = True
            # Avoid RuntimeError: dictionary changed size during iteration
            BaseConnectionHandler.__iter__ = lambda s: iter(dict(s.settings))
            request._rid = str(id(request))
            settings.DATABASES[request._rid] = settings.DATABASES[
                connection._alias
            ]
            rid.set(request._rid)

        request._start_time = monotonic_ns()

    def process_response(self, request, response):
        req_time = (monotonic_ns() - request._start_time) / 1000000

        if request._profile:
            request._profile.disable()
            s = io.StringIO()
            sortby = pstats.SortKey.CUMULATIVE
            ps = pstats.Stats(request._profile, stream=s).sort_stats(sortby)
            ps.print_stats()

        if request._db_router:
            queries = connections[request._rid].queries[:]
        else:
            queries = connection.queries[:]
        sqltime = 0
        sqls = []
        for query in queries:
            sqltime += float(query["time"]) * 1000
            sqlquery = sqlparse.format(
                query['sql'], reindent=True, wrap_after=160)
            if settings.PFX_PROFILING_SQL and request.path.startswith(
                    settings.PFX_PROFILING_SQL):
                sqls.append(f"\n---\n{sqlquery}")
        if sqls:
            sqls.append("\n---")

        sqltime_p = 100 * sqltime / req_time
        logger.debug(
            f"\n    {request.method} {request.get_full_path()} "
            f"{response.status_code} [{req_time:.0f}ms]"
            f"\n    SQL queries: {len(queries)} "
            f"[{sqltime:.0f}ms {sqltime_p:.2f}%]"
            f"{''.join(sqls)}"
            + (request._profile and f"\n{s.getvalue()}" or ""))

        if request._db_router:
            connections[request._rid].close()
            settings.DATABASES.pop(request._rid)
        return response



import logging
from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth.hashers import make_password
from django.test import TransactionTestCase, override_settings

import jwt
from freezegun import freeze_time

from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests_custom_user.models import User

logger = logging.getLogger(__name__)


class AuthAPITest(TestAssertMixin, TransactionTestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')
        self.cookie_client = APIClient(default_locale='en', with_cookie=True)
        self.user1 = User.objects.create(
            username='jrr.tolkien',
            password=make_password('RIGHT PASSWORD'),
        )

    def test_invalid_login(self):
        response = self.client.post(
            '/api/auth/login', {
                'username': 'jrr.tolkien',
                'password': 'WRONG PASSWORD'})
        self.assertRC(response, 422)

    def test_emtpy_login(self):
        response = self.client.post(
            '/api/auth/login', {})
        self.assertRC(response, 422)

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_valid_login(self):
        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'RIGHT PASSWORD'})

            self.assertRC(response, 200)
            token = self.get_val(response, 'token')
            headers = jwt.get_unverified_header(token)
            self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
            decoded = jwt.decode(
                token, self.user1.password + settings.PFX_SECRET_KEY,
                algorithms="HS256")
            self.assertTrue(
                datetime.utcnow() + timedelta(minutes=25) <
                datetime.utcfromtimestamp(decoded['exp'])
                < datetime.utcnow() + timedelta(minutes=35))



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import logging
import os
import sys

from django.db.migrations import writer

if len(sys.argv) > 1 and sys.argv[1] == 'test':
    logging.disable(logging.WARNING)


writer.MIGRATION_HEADER_TEMPLATE = """\
# Generated by Django %(version)s on %(timestamp)s
# flake8: noqa

"""

def main():
    """Run administrative tasks."""
    mode = os.environ.get('CI', '').lower() == 'true' and 'ci' or 'dev'
    argv = [a for a in sys.argv if not a.startswith('--')]
    options = [a for a in sys.argv if a.startswith('--')]
    settings = f"pfx.settings.dev"
    cmd = len(argv) > 1 and argv[1]
    if cmd == 'test':
        app = len(argv) > 2 and argv[2] or 'tests'
        if len(argv) <= 2:
            argv.append('tests')
        settings = f"{app.split('.')[0]}.settings.{mode}"
    elif cmd == 'makemigrations':
        app = len(argv) > 2 and argv[2] or 'pfx'
        settings = f"{app}.settings.{app == 'pfx' and 'dev' or 'common'}"
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings)
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    execute_from_command_line([*argv, *options])


if __name__ == '__main__':
    main()



import json
import logging
import re
from json import JSONDecodeError

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import FieldError, ValidationError
from django.db import IntegrityError, connection, transaction
from django.db.models import F, ForeignKey, ManyToOneRel, Model, Q, Window
from django.db.models.fields import AutoFieldMixin
from django.db.models.functions import Lag, Lead, RowNumber
from django.urls import path
from django.utils.translation import gettext_lazy as _
from django.views import View

import dill

from pfx.pfxcore import __PFX_VIEWS__
from pfx.pfxcore.apidoc import ModelListSchema, ModelSchema, Schema, Tag
from pfx.pfxcore.decorator import rest_api
from pfx.pfxcore.exceptions import (
    APIError,
    ForbiddenError,
    JsonErrorAPIError,
    ModelValidationAPIError,
    NotFoundError,
    UnauthorizedError,
)
from pfx.pfxcore.http import JsonResponse
from pfx.pfxcore.models import (
    JSONReprMixin,
    OrderedModelMixin,
    UserFilteredQuerySetMixin,
)
from pfx.pfxcore.shortcuts import (
    class_key,
    f,
    get_bool,
    get_int,
    get_object,
    model_permissions,
)
from pfx.pfxcore.views.fields import ModelList

from . import parameters
from .fields import FieldType as FT
from .fields import process_fields

logger = logging.getLogger(__name__)


# HTTP 404 handler
def resource_not_found(request, exception):
    return NotFoundError().response


class ModelMixin():
    """Base mixin for a model view."""

    #: The model class, must be defined on concrete classes.
    model = None
    #: The default fields list.
    fields = []

    def apply_user_filter(self, qs):
        """Apply filters to restrict the queryset according to user rights.

        :param qs: The queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The filtered queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        if isinstance(qs, UserFilteredQuerySetMixin):
            return qs.user(self.request.user)
        if (hasattr(settings, 'PFX_FORCE_USER_FILTERED_QUERYSET') and
                settings.PFX_FORCE_USER_FILTERED_QUERYSET):
            raise Exception("The queryset must be a UserFilteredQuerySetMixin")
        return qs

    def get_queryset(
            self, select_related=None, prefetch_related=None,
            from_queryset=None):
        """Get the queryset for the view model.

        The returned queryset is filtered according to user rights.

        :param select_related: Arguments for queryset select_related
        :type select_related: :class:`list[str]`
        :param prefetch_related: Arguments for queryset prefetch_related
        :type prefetch_related: :class:`list[str]`
        :param from_queryset: A source queryset
        :returns: The queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        qs = self.apply_user_filter(
            from_queryset or self.model._default_manager.all())
        if select_related:
            qs = qs.select_related(*select_related)
        if prefetch_related:
            qs = qs.prefetch_related(*prefetch_related)
        return qs

    def get_object(self, **kwargs):
        """Get a single object matching the given keyword arguments.

        :param \\**kwargs: Keyword arguments
        :returns: A view model instance
        """
        return get_object(self.get_queryset(
            select_related=self.get_fields_select_related(),
            prefetch_related=self.get_fields_prefetch_related()), **kwargs)

    def get_related_queryset(
            self, related_model, select_related=None, prefetch_related=None):
        """Get a queryset for another related model.

        The returned queryset is filtered according to user rights.

        :param related_model: An arbitrary DJango model.
        :param select_related: Arguments for queryset select_related
        :type select_related: :class:`list[str]`
        :param prefetch_related: Arguments for queryset prefetch_related
        :type prefetch_related: :class:`list[str]`
        :returns: The queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        qs = self.apply_user_filter(related_model._default_manager.all())
        if select_related:
            qs = qs.select_related(*select_related)
        if prefetch_related:
            qs = qs.prefetch_related(*prefetch_related)
        return qs

    def get_list_queryset(self, select_related=None, prefetch_related=None):
        """Get a queryset for the view model dedicated for lists.

        This method returns the same result as :code:`get_queryset` and is
        defined only to allow it to be overloaded in order to customize
        the queryset for lists.

        :param select_related: Arguments for queryset select_related
        :type select_related: :class:`list[str]`
        :param prefetch_related: Arguments for queryset prefetch_related
        :type prefetch_related: :class:`list[str]`
        :returns: The queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        return self.get_queryset(select_related, prefetch_related)

    @classmethod
    def _process_fields(cls, fields):
        return process_fields(cls.model, fields)

    @classmethod
    def get_fields(cls):
        """Return the processed field list.

        Return a list of :class:`ViewField` built from :code:`cls.fields`.

        The result is cached.

        :returns: The fields list
        :rtype: :class:`list[pfx.pfxcore.views.ViewField]`
        """
        return dill.loads(cache.get_or_set(
            class_key(cls, 'fields'),
            lambda: dill.dumps(cls._process_fields(cls.fields)), None))

    @classmethod
    def get_fields_select_related(cls):
        """:meta private: Undocumented because it must be changed."""
        return cache.get_or_set(
            class_key(cls, 'fields', 'select_related'),
            lambda: set([
                _f for field in cls.get_fields().values()
                for _f in field.select_related]),
            None)

    @classmethod
    def get_fields_prefetch_related(cls):
        """:meta private: Undocumented because it must be changed."""
        return cache.get_or_set(
            class_key(cls, 'fields', 'prefetch_related'),
            lambda: set([
                _f for field in cls.get_fields().values()
                for _f in field.prefetch_related]),
            None)

    @property
    def model_name(self):
        """Get the model verbose name.

        :returns: The model
        :rtype: :class:`str`
        """
        return self.model._meta.verbose_name

    def message_response(self, message, **kwargs):
        """Build a message JSON response.

        :param message: A message
        :type message: str
        :param \\**kwargs: Other JSON values to add to the response.
        :returns: The response
        :rtype: :class:`pfx.pfxcore.http.JsonResponse`
        """
        return JsonResponse(dict(message=message, **kwargs))

    def delete_object(self, obj):
        """Delete a model object instance.

        Raise an :class:`APIError` for :class:`IntegrityError`.
        If the exception is not catch,
        A JSON HTTP response is returned by the API call.

        :param obj: The object to delete
        """
        try:
            with transaction.atomic():
                obj.delete()
        except IntegrityError as e:
            logger.debug("IntegrityError: %s", e)
            raise APIError(f(_(
                "{obj} cannot be deleted because "
                "it is referenced by other objects."), obj=obj))

    @classmethod
    def get_apidoc_tags(cls):
        """Get the tags for ApiDoc.
        """
        return cls.tags or [
            Tag(str(cls.model._meta.verbose_name))]

    @classmethod
    def get_model_perms(cls, actions):
        """Get the model permission name for the action.
        """
        return model_permissions(cls.model, actions)

    def pre_save(self, obj, created=False):
        """Return a list of function to call in post_save.
        """
        return []

    def post_save(self, obj, created=False, funcs=None):
        """Post save operations.
        """
        funcs = funcs or []
        for func in funcs:
            func()


class ModelResponseMixin(ModelMixin):
    """Extension of :class:`ModelMixin` to manage object responses."""
    #: The response schema for meta service.
    meta_schema = None
    #: The model schema.
    model_schema = None
    #: The create response schema.
    model_create_schema = None
    #: The update response schema.
    model_update_schema = None
    #: The message response schema.
    model_message_schema = None

    def serialize_object(self, obj, **fields):
        """Serialize an object into a python :class:`dict`.

        If the object is a :class:`JSONReprMixin`,
        use the :code:`json_repr` method.
        Otherwise return a default representation.

        :param obj: The object
        :param \\**fields: Other values to add to the representation
        :returns: The object representation
        :rtype: :class:`dict`
        """
        if isinstance(obj, JSONReprMixin):
            vals = obj.json_repr()
        else:
            vals = dict(
                pk=obj.pk,
                resource_name=str(obj))
        vals.update(fields)
        return vals

    def get_navigation_meta(self, obj):
        return {}

    def response(self, o, **meta):
        """Build a :class:`JsonResponse` from an object instance.

        :param o: The object instance
        :param \\**meta: Meta values to add to the response
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        """
        if get_bool(self.request.GET, 'navigation'):
            meta.update(self.get_navigation_meta(o))
        return JsonResponse(self.serialize_object(o, **{
            _f.alias: _f.to_json(o, self)
            for _f in self.get_fields().values()}, meta=meta))

    def validate(self, obj, rel_data=None, created=False, **kwargs):
        """Validate an object instance.

        can be overridden to customize validation on a given view
        (apart from object validation).

        :param o: The object instance to validate
        :param created: If object instance is created
        :param \\**kwargs: Additional arguments for :code:`full_clean`
        """
        def model_lists(rel_data):
            if rel_data:
                for k, v in rel_data.items():
                    if isinstance(v, ModelList):
                        yield k, v

        errors = {}
        for k, v in model_lists(rel_data):
            for i, o in enumerate(v):
                save_field = getattr(o, '_save_related', False)
                if save_field:
                    try:
                        setattr(o, save_field, obj)
                        o.full_clean(exclude={save_field})
                    except ValidationError as e:
                        for mk, ms in e.error_dict.items():
                            errors.setdefault(
                                f'{k}::{i}::{mk}', []).extend(ms)
        obj._rel_data = rel_data or {}
        if errors:
            try:
                obj._rel_data = rel_data
                obj.full_clean(**kwargs)
            except ValidationError as e:
                for k, ms in e.error_dict.items():
                    errors.setdefault(k, []).extend(ms)
            raise ValidationError(errors)
        else:
            obj.full_clean(**kwargs)

    def is_valid_response_meta(self, obj, created=False):
        """Prepare the defaut meta for is_valid responce.

        :param obj: The object instance
        :param created: If object instance is created
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        """
        message = (
                created and
                _("{model} {obj} created.") or
                _("{model} {obj} updated."))
        return dict(
            created=created,
            message=f(
                message, model=self.model_name, obj=object))

    def is_valid(self, obj, created=False, rel_data=None):
        """Persist an object instance changes and build default response.

        The default response contains the serialized instance after save and
        a text message.

        You can use :code:`rel_data` to pass values to set on related fields
        after instance is persisted (to avoid errors if the instance does
        not exists in database).

        :param obj: The object instance
        :param created: If object instance is created
        :param rel_data: Values to set on related fields
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        """
        with transaction.atomic():
            funcs = self.pre_save(obj, created=created)
            obj.save()
            if rel_data:
                for k, v in rel_data.items():
                    if isinstance(v, ModelList):
                        if isinstance(obj._meta.get_field(k), ManyToOneRel):
                            for cur in getattr(obj, k).all():
                                if cur not in v:
                                    try:
                                        cur.delete()
                                    except IntegrityError:
                                        raise APIError(f(_(
                                            "{obj} cannot be deleted because "
                                            "it is referenced by other "
                                            "objects."), obj=str(cur)))
                        prev = None
                        for o in v:
                            save_field = getattr(o, '_save_related', False)
                            if save_field:
                                setattr(o, save_field, obj)
                                funcs += self.pre_save(o, created=created)
                                o.save()
                                if isinstance(o, OrderedModelMixin) and prev:
                                    o.below(prev)
                            prev = o
                    getattr(obj, k).set(v)
            self.post_save(obj, created=created, funcs=funcs)
        obj = self.get_object(pk=obj.pk)
        return self.response(
            obj, **self.is_valid_response_meta(obj, created=created))

    def is_invalid(self, obj, errors):
        """Build a 422 response for invalid object instance.

        :param obj: The object instance
        :param errors: The validations errors
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        """
        return JsonResponse(errors, status=422)

    def object_meta(self):
        """Build metadata for the view model.

        :returns: The metadata
        :rtype: :class:`dict`
        """
        meta = {}
        defaults = {} if self.request.GET.keys() else dict(
            fields=True, model=True)
        if get_bool(self.request.GET, 'fields', defaults.get('fields')):
            meta['fields'] = cache.get_or_set(
                class_key(self.__class__, 'meta', 'fields'),
                lambda: {n: f.meta() for n, f in self.get_fields().items()},
                None)
        if get_bool(self.request.GET, 'model', defaults.get('model')):
            meta['model'] = {
                'app': self.model._meta.app_label,
                'name': self.model._meta.model_name,
                'object': self.model._meta.object_name,
            }
        return meta

    @rest_api(
        "/meta", method="get", response_schema='meta_schema', priority_doc=20)
    def get_meta(self, *args, **kwargs):
        """Entrypoint for :code:`GET /meta` route.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get {model} metadata
        """
        return JsonResponse(self.object_meta())

    @classmethod
    def generate_schemas(cls):
        """Generate schemas for the class.
        """
        from .fields import FieldType
        super().generate_schemas()
        cls.meta_schema = Schema('form_meta', "Meta", properties=dict(
            fields=dict(
                type='array', items=dict(type='object', properties=dict(
                    name=dict(type='string'),
                    type=dict(
                        type='string',
                        enum=list(FieldType.APIDOC_FIELD_BINDING.keys())),
                    choices=dict(
                        type='array', items=dict(type='string'),
                        example=["value1", "value2"]),
                    readonly=dict(type='object', properties=dict(
                        post=dict(type='boolean'),
                        put=dict(type='boolean'))),
                    required=dict(type='boolean'))),
                description="List of fields.")))
        cls.model_schema = ModelSchema(cls.model, cls.get_fields())
        cls.model_create_schema = ModelSchema(
            cls.model, cls.get_fields(), mode='create')
        cls.model_update_schema = ModelSchema(
            cls.model, cls.get_fields(), mode='update')
        cls.model_message_schema = ModelSchema(
            cls.model, cls.get_fields(), dict(message=dict(type='string')))

    @classmethod
    def get_apidoc_schemas(cls):
        """Get schemas for the class.

        :returns: The schemas list
        :rtype: :class:`list[pfx.pfxcore.apidoc.Schema]`
        """
        return super().get_apidoc_schemas() + [
            cls.meta_schema, cls.model_schema, cls.model_create_schema,
            cls.model_update_schema, cls.model_message_schema]


class BodyMixin:
    """Base mixin for request body management."""

    def deserialize_body(self):
        """Return the request body as a python :class:`dict`.
        """
        try:
            return json.loads(self.request.body)
        except JSONDecodeError as e:
            raise JsonErrorAPIError(e)

    def body_to_model(
            self, model: Model, validate=True, fields=None, **kwargs):
        """Return a new model instance built with request body.

        :param model: The model to instantiate
        :param validate: Activate validation
        :param \\**kwargs: The optional arguments for `full_clean`
        :rtype: :class:`django.db.models.Model`
        """
        if fields is None:
            fields = [
                _f.name for _f in model._meta.get_fields()
                if not isinstance(_f, AutoFieldMixin)]
        obj = model(**{
            k: v for k, v in self.deserialize_body().items() if k in fields})
        if validate:
            try:
                obj.full_clean(**kwargs)
            except ValidationError as e:
                raise ModelValidationAPIError(e)
        return obj


class ModelBodyMixin(BodyMixin, ModelMixin):
    """Extension mixin to process object in body."""

    def get_model_data(self, obj, data, created):
        """Process data for object update or creation.

        Return a tuple of data and rel_data. The returned data contains
        only fields that can be modified in this context (create or update).

        Values are deserialized according to field definition. If the field
        il a related object list, the value is set in rel_data.

        :returns: The data and rel_data
        :rtype: :class:`tuple(dict, dict)`
        """
        fields = self.get_fields()

        def can_write(fname, fields):
            if fname not in fields:
                return False
            if fields[fname].is_readonly(created=created):
                logger.warning(
                    "Field %s is ignored because it is readonly on view %s",
                    fname, self.__class__.__name__)
                return False
            return True

        res = {}
        res_rel = {}
        for k, v in data.items():
            if can_write(k, fields):
                field = fields[k]
                mk, mv = field.to_model_value(v, self.get_related_queryset)
                if field.field_type == FT.ModelObjectList:
                    if field.related_fields:
                        for i, rv in enumerate(v):
                            for rk, rf in field.related_fields.items():
                                rmk, rmv = rf.to_model_value(
                                    rv.get(rk), self.get_related_queryset)
                                setattr(mv[i], rmk, rmv)
                            setattr(
                                mv[i], '_save_related',
                                field.model_field.field.name)
                    res_rel[mk] = mv
                else:
                    res[mk] = mv
        return res, res_rel

    def set_values(self, obj, **values):
        """Set object fields value.

        :param obj: The object to update
        :param \\**values: The values to set
        """
        for fname, value in values.items():
            setattr(obj, fname, value)


class ListRestViewMixin(ModelResponseMixin):
    """Extension mixin to manage object list response."""

    class Subset:
        """Subset enum."""
        NONE = 'none'
        PAGINATION = 'pagination'
        OFFSET = 'offset'

    #: The fields for list responses.
    list_fields = []
    #: The filters available for list responses.
    filters = []
    #: The default order (model default is used if empty).
    default_order = []
    #: The schema for list metadata.
    meta_list_schema = None
    #: The schema for responses.
    model_list_schema = None

    def get_list_fields(self):
        """Return the processed field list for lists.

        Return a list of :class:`ViewField` built from :code:`cls.list_fields`
        (or :code:`cls.fields` if :code:`cls.list_fields` is empty).

        the result is cached on the class, to ensure that the build
        is carried out no more than once per request.

        :returns: The fields list
        :rtype: :class:`list[pfx.pfxcore.views.ViewField]`
        """
        if not hasattr(self, '_list_fields'):
            self._list_fields = dill.loads(cache.get_or_set(
                class_key(self.__class__, 'list_fields'),
                lambda: dill.dumps(self._process_fields(
                    self.list_fields or self.fields)), None))
        return self._list_fields

    def get_list_meta_filters(self):
        """Return the filters metadata for lists.

        :returns: The filters metadata generator
        :rtype: :class:`generator`
        """
        for _f in self.filters:
            yield _f.meta

    def search_filter(self, search):  # pragma: no cover
        """Return the django filters for the default text search.

        This default implementation returns empty filters, which will always
        produce empty results. This method is designed to be implemented
        by the view to be customized.

        :returns: The django filters
        :rtype: :class:`django.db.models.Q`
        """
        if hasattr(self.model.objects, 'default_search'):
            return self.model.objects.default_search(search)
        return Q()

    def orderable_fields(self, model, models=None):
        """:meta private: Undocumented because it must be changed/removed."""
        models = models or [model]
        new_models = models + [
            field.related_model for field in model._meta.fields
            if isinstance(field, ForeignKey)]
        for field in model._meta.fields:
            if isinstance(field, ForeignKey):
                if field.related_model not in models:
                    yield field.name
                    yield from [
                        f"{field.name}__{fn}" for fn in self.orderable_fields(
                            field.related_model, new_models)]
                continue
            else:
                yield field.name == 'id' and 'pk' or field.name

    def object_meta_list(self):
        """Build metadata for the view model list.

        :returns: The metadata
        :rtype: :class:`dict`
        """
        default_all = not self.request.GET.keys()
        meta = {}
        if get_bool(self.request.GET, 'fields', default_all):
            meta['fields'] = cache.get_or_set(
                class_key(self.__class__, 'meta', 'list_fields'),
                lambda: {
                    n: f.meta() for n, f in self.get_list_fields().items()},
                None)
        if get_bool(self.request.GET, 'filters', default_all):
            meta['filters'] = cache.get_or_set(
                class_key(self.__class__, 'meta', 'filters'),
                lambda: [_f.meta for _f in self.filters],
                None)
        if get_bool(self.request.GET, 'orders'):
            meta['orders'] = cache.get_or_set(
                class_key(self.__class__, 'meta', 'orders'),
                lambda: list(self.orderable_fields(self.model)),
                None)
        if get_bool(self.request.GET, 'model', default_all):
            meta['model'] = {
                'app': self.model._meta.app_label,
                'name': self.model._meta.model_name,
                'object': self.model._meta.object_name,
            }
        return meta

    @rest_api("/meta/list", method="get", parameters=[
        parameters.groups.MetaList], response_schema='meta_list_schema',
        priority_doc=10)
    def get_meta_list(self, *args, **kwargs):
        """Entrypoint for :code:`GET /meta/list` route.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get {models} list metadata
        """
        return JsonResponse(self.object_meta_list())

    def apply_view_filter(self, qs):
        """Apply view filters on queryset.

        :param qs: The source queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The filtered queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        for filter in self.filters:
            qs = qs.filter(filter.query(self.request.GET))
        return qs

    def apply_view_search(self, qs):
        """Apply view search on queryset.

        :param qs: The source queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The filtered queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        q = None
        for search in self.request.GET.getlist('search'):
            crit = self.search_filter(search)
            if q is None:
                q = crit
            q |= crit
        if q:
            return qs.filter(q)
        return qs

    def get_order_mapping(self):
        return {}

    def get_query_order(self):
        omap = self.get_order_mapping()

        def apply(o):
            if not o:
                return o
            if o[0] == '+':
                o = o[1:]
            s, fn = o[0] in '-' and ('-', o[1:]) or ('', o)
            return f"{s}{omap.get(fn, fn)}"

        order = self.request.GET.get('order')
        return order and [apply(o) for o in order.split(',')] or order

    def get_order(self):
        order = (
            self.get_query_order() or self.default_order or
            self.model._meta.ordering)
        if 'pk' not in order:
            order.append('pk')
        return order

    def apply_view_order(self, qs):
        """Apply view order on queryset.

        :param qs: The source queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The ordered queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        try:
            return qs.order_by(*self.get_order())
        except FieldError as e:
            raise APIError(e.args[0])

    def get_list_queryset(self, select_related=None, prefetch_related=None):
        """Get list queryset.

        The result queryset is filtered and ordered by view settings.

        :param select_related: Arguments for queryset select_related
        :type select_related: :class:`list[str]`
        :param prefetch_related: Arguments for queryset prefetch_related
        :type prefetch_related: :class:`list[str]`
        :returns: The filtered queryset
        :rtype: :class:`django.db.models.QuerySet`
        """
        qs = super().get_list_queryset(select_related, prefetch_related)
        qs = self.apply_view_filter(qs)
        qs = self.apply_view_search(qs)
        qs = self.apply_view_order(qs)
        return qs.distinct()

    def get_list_fields_select_related(self):
        """:meta private: Undocumented because it must be changed."""
        return cache.get_or_set(
            class_key(self.__class__, 'list_fields', 'select_related'),
            lambda: set([
                _f for field in self.get_list_fields().values()
                for _f in field.select_related]),
            None)

    def get_list_fields_prefetch_related(self):
        """:meta private: Undocumented because it must be changed."""
        return cache.get_or_set(
            class_key(self.__class__, 'list_fields', 'prefetch_related'),
            lambda: set([
                _f for field in self.get_list_fields().values()
                for _f in field.prefetch_related]),
            None)

    def get_list_result(self, qs):
        """Get a generator to serialize each result in a queryset.

        :param qs: The source queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The generator
        :rtype: :class:`generator`
        """
        qs = qs.select_related(
            *self.get_list_fields_select_related()
        ).prefetch_related(
            *self.get_list_fields_prefetch_related())
        for o in qs:
            yield self.serialize_object(o, **{
                _f.alias: _f.to_json(o, self)
                for _f in self.get_list_fields().values()})

    def get_short_list_result(self, qs):
        """Get a generator to serialize each result in a queryset.

        Each result is serialized with its simple representation (json_repr
        od default representation).

        :param qs: The source queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The generator
        :rtype: :class:`generator`
        """
        for o in qs:
            if isinstance(o, JSONReprMixin):
                yield o.json_repr()
            else:
                yield dict(
                    pk=o.pk,
                    resource_name=str(o))

    def pagination_count_queryset(self, qs):
        return qs

    def pagination_apply(self, qs, count_qs, offset, page_size):
        return qs.all()[offset:offset + page_size]

    def pagination_result(self, qs):
        """Apply pagination on a queryset.

        :param qs: The source queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The queryset for the active page and the pagination data
        :rtype: :class:`tuple(django.db.models.QuerySet, dict)`
        """
        page_size = get_int(self.request.GET, 'page_size', 10)
        if settings.PFX_MAX_LIST_RESULT_SIZE:
            page_size = min(page_size, settings.PFX_MAX_LIST_RESULT_SIZE)
        page = get_int(self.request.GET, 'page', 1)
        page_subset = get_int(self.request.GET, 'page_subset', 5)
        cqs = self.pagination_count_queryset(qs)
        count = cqs.count()
        page_count = (1 + (count - 1) // page_size) or 1
        offset = (page - 1) * page_size
        subset_first = min(
            max(page - page_subset // 2, 1),
            max(page_count - page_subset + 1, 1))
        qs = self.pagination_apply(qs, cqs, offset, page_size)
        return qs, dict(
            page_size=page_size,
            page=page,
            page_subset=page_subset,
            count=count,
            page_count=page_count,
            subset=list(range(
                subset_first,
                min(subset_first + page_subset, page_count + 1))))

    def offset_result(self, qs):
        """Apply offset/limit on a queryset.

        :param qs: The source queryset
        :type qs: :class:`django.db.models.QuerySet`
        :returns: The queryset for the active page and the pagination data
        :rtype: :class:`tuple(django.db.models.QuerySet, dict)`
        """
        limit = get_int(self.request.GET, 'limit', 10)
        if settings.PFX_MAX_LIST_RESULT_SIZE:
            limit = min(limit, settings.PFX_MAX_LIST_RESULT_SIZE)
        offset = get_int(self.request.GET, 'offset', 0)
        count = qs.count()
        page_count = (1 + (count - 1) // limit) or 1
        qs = qs[offset:offset + limit]
        return qs, dict(
            count=count,
            page_count=page_count,
            limit=limit,
            offset=offset)

    def _get_list_extra_meta(self, res):
        return {}

    def _get_list(self, *args, **kwargs):
        res = {}
        meta = {}
        qs = self.get_list_queryset()
        subset = self.request.GET.get('subset', self.Subset.NONE)
        if get_bool(self.request.GET, 'count'):
            meta['count'] = qs.count()
        if subset == self.Subset.PAGINATION:
            qs, meta['subset'] = self.pagination_result(qs)
        elif subset == self.Subset.OFFSET:
            qs, meta['subset'] = self.offset_result(qs)
        elif settings.PFX_MAX_LIST_RESULT_SIZE:
            qs = qs[:settings.PFX_MAX_LIST_RESULT_SIZE]
        if get_bool(self.request.GET, 'items', 'count' not in meta):
            mode = self.request.GET.get('mode', 'list')
            match mode:
                case 'list':
                    res['items'] = list(self.get_list_result(qs))
                case 'select':
                    res['items'] = list(self.get_short_list_result(qs))
                case _:
                    m = f'get_{mode}_result'
                    res['items'] = list(getattr(self, m)(qs))
        meta.update(self._get_list_extra_meta(res))
        if meta:
            res['meta'] = meta
        return JsonResponse(res)

    @rest_api(
        "", method="get", parameters=[parameters.groups.List],
        filters=True, search=True,
        response_schema='model_list_schema', priority_doc=-20)
    def get_list(self, *args, **kwargs):
        """Entrypoint for :code:`GET /` route.

        Retrieve an object list response.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get {models} list
        """

        return self._get_list(*args, **kwargs)

    def get_navigation_meta(self, obj):
        qs = self.get_list_queryset()
        order = qs.query.order_by
        qs_window = qs.annotate(
            prev_id=Window(Lag('id'), order_by=order),
            next_id=Window(Lead('id'), order_by=order),
            index=Window(RowNumber(), order_by=order)).values(
                'prev_id', 'next_id', 'index', obj_id=F('id'))
        compiler = qs_window.query.get_compiler(connection=connection)
        inner_sql, params = compiler.as_sql()
        sql = f"""
            WITH annotated AS ({inner_sql})
            SELECT prev_id, next_id, index
            FROM annotated
            WHERE obj_id = %s;
        """

        with connection.cursor() as cursor:
            cursor.execute(sql, list(params) + [obj.pk])
            prev_id, next_id, index = cursor.fetchone() or (None, None)
            return {
                'previous_pk': prev_id or qs.last().pk,
                'next_pk': next_id or qs.first().pk,
                'index': index,
                'count': qs.count()}

    @classmethod
    def generate_schemas(cls):
        """Generate schemas for the class.
        """
        from .fields import FieldType
        super().generate_schemas()
        cls.meta_list_schema = Schema('list_meta', "Meta", properties=dict(
            fields=dict(
                type='array', items=dict(type='object', properties=dict(
                    name=dict(type='string'),
                    type=dict(
                        type='string',
                        enum=list(FieldType.APIDOC_FIELD_BINDING.keys())),
                    choices=dict(
                        type='array', items=dict(type='string'),
                        example=["value1", "value2"]),
                    readonly=dict(type='object', properties=dict(
                        post=dict(type='boolean'),
                        put=dict(type='boolean'))),
                    required=dict(type='boolean'))),
                description="List of fields."),
            filters=dict(
                type='array', items=dict(type='object', properties=dict(
                    name=dict(type='string'),
                    label=dict(type='string'),
                    type=dict(
                        type='string',
                        enum=list(FieldType.APIDOC_FIELD_BINDING.keys())),
                    choices=dict(
                        type='array', items=dict(type='string'),
                        example=["value1", "value2"]),
                    empty_value=dict(type='boolean'),
                    technical=dict(type='boolean'),
                    related_model=dict(type='string'))),
                description="List of fields."),
            orders=dict(
                type='array', items=dict(type='string'),
                description="List of orderable fields.",
                example=['field1', 'field2', 'field3'])))
        cls.model_list_schema = ModelListSchema(
            cls.model, cls._process_fields(cls.list_fields or cls.fields))

    @classmethod
    def get_apidoc_schemas(cls):
        """Get schemas for the class.

        :returns: The schemas list
        :rtype: :class:`list[pfx.pfxcore.apidoc.Schema]`
        """
        return super().get_apidoc_schemas() + [
            cls.meta_list_schema, cls.model_list_schema]


class ListPermsRestViewMixin(ListRestViewMixin):
    """Extension mixin to check permissions."""

    def get_list_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('view'))


class DetailRestViewMixin(ModelResponseMixin):
    """Extension mixin to add a get detail route."""

    @rest_api("/<int:id>", method="get", parameters=[
        parameters.groups.ModelSerialization], response_schema='model_schema',
        priority_doc=-10)
    def get(self, id, *args, **kwargs):
        """Entrypoint for :code:`GET /<int:id>` route.

        Retrieve an object detail by ID.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get {model}
            parameters extras:
                id: the {model} pk
        """
        obj = self.get_object(pk=id)
        return self.response(obj)


class DetailPermsRestViewMixin(DetailRestViewMixin):
    """Extension mixin to check permissions."""

    def get_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('view'))


class SlugDetailRestViewMixin(ModelResponseMixin):
    """Extension mixin to add a get detail by slug route."""

    #: The slug field name
    SLUG_FIELD = "slug"

    @rest_api(
        "/slug/<slug:slug>", method="get",
        parameters=[parameters.groups.ModelSerialization],
        response_schema='model_schema', priority_doc=-10)
    def get_by_slug(self, slug, *args, **kwargs):
        """Entrypoint for :code:`GET /slug/<slug:slug>` route.

        Retrieve an object detail by slug.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get {model} by slug
            parameters extras:
                slug: the {model} slug name
        """
        obj = self.get_object(**{self.SLUG_FIELD: slug})
        return self.response(obj)


class SlugPermsDetailRestViewMixin(SlugDetailRestViewMixin):
    """Extension mixin to check permissions."""

    def get_by_slug_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('view'))


class CreateRestViewMixin(ModelBodyMixin, ModelResponseMixin):
    """Extension mixin to add create route."""

    #: Default values
    default_values = {}

    def get_default_values(self):
        """Get default values

        :returns: The default values
        :rtype: :class:`dict`
        """
        return dict(self.default_values)

    def new_object(self):
        """Get a new object instance with default values.

        :returns: A model instance
        :rtype: :class:`django.db.models.Model`
        """
        return self.model(**self.get_default_values())

    def object_create_perm(self, data):
        """Return :code:`True` if an instance can be created.

        This method returns always :code:`True`. It is designed to be
        overloaded to customize permissions. You can customize depending
        on user and/or creation data.

        :param data: The data used to create instance
        :returns: A model instance
        :rtype: :class:`django.db.models.Model`
        """
        return True

    def _post(self, *args, **kwargs):
        try:
            obj = self.new_object()
            data, rel_data = self.get_model_data(
                obj, self.deserialize_body(), created=True)
            forbidden = False
            if not self.object_create_perm(data):
                forbidden = True
            self.set_values(obj, **data)
            self.validate(obj, rel_data=rel_data, created=True)
            if forbidden:
                raise ForbiddenError
            return self.is_valid(obj, created=True, rel_data=rel_data)
        except ValidationError as e:
            return self.is_invalid(obj, errors=e)

    @rest_api(
        "", method="post", parameters=[parameters.groups.ModelSerialization],
        request_schema='model_create_schema',
        response_schema='model_message_schema', priority_doc=-20)
    def post(self, *args, **kwargs):
        """Entrypoint for :code:`POST /` route.

        Create an object from JSON body.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        post:
            summary: Create {model}
        """
        return self._post(*args, **kwargs)


class CreatePermsRestViewMixin(CreateRestViewMixin):
    """Extension mixin to check permissions."""

    def post_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('add'))


class UpdateRestViewMixin(ModelBodyMixin, ModelResponseMixin):
    """Extension mixin to add create route."""

    def object_update_perm(self, obj, data):
        """Return :code:`True` if an instance can be updated.

        This method returns always :code:`True`. It is designed to be
        overloaded to customize permissions. You can customize depending
        on user, current instance and/or updated data.

        :param obj: Current instance
        :param data: The data used to create instance
        :returns: A model instance
        :rtype: :class:`django.db.models.Model`
        """
        return True

    def _put(self, id, *args, **kwargs):
        try:
            obj = self.get_object(pk=id)
            data, rel_data = self.get_model_data(
                obj, self.deserialize_body(), created=False)
            forbidden = False
            if not self.object_update_perm(obj, data):
                forbidden = True
            self.set_values(obj, **data)
            self.validate(obj, rel_data=rel_data, created=False)
            if forbidden:
                raise ForbiddenError
            return self.is_valid(obj, created=False, rel_data=rel_data)
        except ValidationError as e:
            return self.is_invalid(obj, errors=e)

    @rest_api(
        "/<int:id>", method="put",
        parameters=[parameters.groups.ModelSerialization],
        request_schema='model_update_schema',
        response_schema='model_message_schema',
        priority_doc=-10)
    def put(self, id, *args, **kwargs):
        """Entrypoint for :code:`PUT /<int:id>` route.

        Update an object from JSON body.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        put:
            summary: Update {model}
            parameters extras:
                id: the {model} pk
        """
        return self._put(id, *args, **kwargs)


class UpdatePermsRestViewMixin(UpdateRestViewMixin):
    """Extension mixin to check permissions."""

    def put_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('change'))


class DeleteRestViewMixin(ModelMixin):
    """Extension mixin to add delete route."""

    def object_delete_perm(self, obj):
        """Return :code:`True` if an instance can be updated.

        This method returns always :code:`True`. It is designed to be
        overloaded to customize permissions. You can customize depending
        on user and/or current instance.

        :param obj: Current instance
        :returns: A model instance
        :rtype: :class:`django.db.models.Model`
        """
        return True

    def _delete(self, id, *args, **kwargs):
        obj = self.get_object(pk=id)
        if not self.object_delete_perm(obj):
            raise ForbiddenError()
        self.delete_object(obj)
        return self.message_response(f(
            _("{model} {obj} deleted."), model=self.model_name, obj=obj))

    @rest_api(
        "/<int:id>", method="delete", response_schema='message_schema',
        priority_doc=-10)
    def delete(self, id, *args, **kwargs):
        """Entrypoint for :code:`DELETE /<int:id>` route.

        Delete an object from JSON body.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        delete:
            summary: Delete {model}
            parameters extras:
                id: the {model} pk
        """
        return self._delete(id, *args, **kwargs)


class DeletePermsRestViewMixin(DeleteRestViewMixin):
    """Extension mixin to check permissions."""

    def delete_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('delete'))


class SecuredRestViewMixin(View):
    """A view mixin to manage service permissions.

    You can add :code:`${func_name}_public` class attributes to override
    :code:`default_public` value for specifics methods. This attribute has
    precedence over the :code:`public` attribute of the :code:`@rest_api`
    decorator, then you can override the behavior in an inherited view
    without overriding the entrypoint method.

    You can add :code:`${func_name}_perm` method to customize permissions
    checks for a dedicated service.
    """

    #: If :code:`True` the services are public by default.
    default_public = False

    def perm(self):
        """Check default permissions for all services.

        Can be overloaded to customize behavior.

        If this method returns :code:`False`, the access is always denied.
        Otherwise custom checks of the service are performed.
        """
        return True

    def _is_public(self, public, func_name):
        param = f'{func_name}_public'
        if hasattr(self, param):
            return getattr(self, param)
        return self.default_public if public is None else public

    def check_perm(self, public, func_name, perms, *args, **kwargs):
        """Check permissions for a specific service.

        Do all checks for a specific service and raise
        :class:`UnauthorizedError` or :class:`ForbiddenError` if needed.

        :param public: The object pk
        :param field: The field name
        """
        if self._is_public(public, func_name):
            return
        if not self.request.user.is_authenticated:
            raise UnauthorizedError()
        if isinstance(perms, str):
            if not self.request.user.has_perm(perms):
                raise ForbiddenError()
        elif perms:
            if not self.request.user.has_perms(perms):
                raise ForbiddenError()
        if not self.perm():
            raise ForbiddenError()
        fperm = f'{func_name}_perm'
        if hasattr(self, fperm) and not getattr(self, fperm)(*args, **kwargs):
            raise ForbiddenError()


class BaseRestView(SecuredRestViewMixin, View):
    """The base class for REST views."""

    #: :meta private: Internal for as_view binding.
    pfx_methods = None
    #: :meta private: Internal for as_view binding.
    rest_view_path = {}
    #: :meta private: Internal for as_view binding.
    rest_doc = {}
    #: :meta private: Internal for as_view binding.
    rest_doc_priority = {}
    #: Tags for ApiDoc.
    tags = None
    #: Schemas for ApiDoc.
    schemas = []
    #: Message schema for ApiDoc.
    message_schema = None

    def dispatch(self, request, *args, **kwargs):
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.
        if request.method.lower() in self.http_method_names:
            handler = getattr(
                self, self.pfx_methods.get(
                    request.method.lower(), 'http_method_not_allowed'),
                self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)

    @staticmethod
    def _path_order(path, methods):
        def process(e):
            e = re.sub(r'<path:.*>', '!01', e)
            e = re.sub(r'<int:.*>', '!03', e)
            e = re.sub(r'<uuid:.*>', '!04', e)
            e = re.sub(r'<slug:.*>', '!05', e)
            e = re.sub(r'<str:.*>', '!06', e)
            e = re.sub(r'<.*>', '!02', e)
            return e

        return methods.get('priority', 0), *map(
            process, path.lstrip('/').split('/'))

    @staticmethod
    def _path_order_doc(path, methods):
        return methods.get('priority_doc', 0), path.lstrip('/').split('/')

    @classmethod
    def get_urls(cls, as_pattern=False):
        """Generate URLs for the view."""
        def fullpath(p2):
            res = f'{cls.rest_view_path[cls]}{p2}'.lstrip('/')
            return res if as_pattern else f'/{res}'

        paths = {}
        for name in dir(cls):
            m = getattr(cls, name, None)
            if m and callable(m) and hasattr(m, 'rest_api_method'):
                methods = paths.setdefault(m.rest_api_path, dict(
                    priority=0, priority_doc=0))
                methods[m.rest_api_method] = name
                if m.rest_api_priority != 0:
                    if methods['priority'] not in (0, m.rest_api_priority):
                        raise Exception(
                            f"Path {fullpath(m.rest_api_path)}: "
                            "you cannot set different priority for same path")
                    methods['priority'] = m.rest_api_priority

                priority_doc = cls.rest_doc_priority.get(
                    (fullpath(m.rest_api_path), m.rest_api_method),
                    m.rest_api_priority_doc)
                if priority_doc != 0:
                    if methods['priority_doc'] not in (
                            0, m.rest_api_priority_doc):
                        raise Exception(
                            f"Path {fullpath(m.rest_api_path)}: "
                            "you cannot set different priority_doc "
                            "for same path")
                    methods['priority_doc'] = m.rest_api_priority_doc
        return [
            path(fullpath(p), cls.as_view(pfx_methods=ms)) if as_pattern
            else dict(path=fullpath(p), methods={
                k: v for k, v in ms.items()
                if k not in ('priority', 'priority_doc')})
            for p, ms in sorted(
                paths.items(), key=lambda e:
                    cls._path_order(*e) if as_pattern
                    else cls._path_order_doc(*e),
                reverse=as_pattern)]

    @classmethod
    def as_urlpatterns(cls):
        """Get URLs patterns for the view."""
        if cls not in __PFX_VIEWS__:
            __PFX_VIEWS__.append(cls)
        return cls.get_urls(as_pattern=True)

    @classmethod
    def get_apidoc_tags(cls):
        """Get ApiDoc tags."""
        return cls.tags or [Tag(str(cls.__name__))]

    @classmethod
    def generate_schemas(cls):
        """Generate schemas for the class.
        """
        cls.message_schema = Schema('Message', "Message", properties=dict(
            message=dict(type='string')))

    @classmethod
    def get_apidoc_schemas(cls):
        """Get schemas for the class.

        :returns: The schemas list
        :rtype: :class:`list[pfx.pfxcore.apidoc.Schema]`
        """
        return cls.schemas + [cls.message_schema]


class RestView(
        ListRestViewMixin,
        DetailRestViewMixin,
        CreateRestViewMixin,
        UpdateRestViewMixin,
        DeleteRestViewMixin,
        BaseRestView):
    pass


class PermsRestView(
        ListPermsRestViewMixin,
        DetailPermsRestViewMixin,
        CreatePermsRestViewMixin,
        UpdatePermsRestViewMixin,
        DeletePermsRestViewMixin,
        BaseRestView):
    pass



import logging

from django.core.exceptions import FieldDoesNotExist
from django.http import FileResponse, HttpResponse
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.decorator import rest_api
from pfx.pfxcore.exceptions import APIError, NotFoundError
from pfx.pfxcore.fields import MediaField
from pfx.pfxcore.http import JsonResponse
from pfx.pfxcore.shortcuts import get_bool, settings
from pfx.pfxcore.storage.s3_storage import StorageException

from . import parameters
from .rest_views import ModelMixin

logger = logging.getLogger(__name__)


def get_medial_field(model, field):
    try:
        model_field = model._meta.get_field(field)
        if not isinstance(model_field, MediaField):
            raise NotFoundError  # pragma: no cover
    except FieldDoesNotExist:  # pragma: no cover
        raise NotFoundError
    return model_field


def get_media_field_response(obj, field, request):
    mediaField = get_medial_field(obj, field)
    try:
        url = mediaField.get_url(request, obj)
    except StorageException as e:  # pragma: no cover
        logger.exception(e)
        raise APIError(_("Unexpected storage error", status=500))

    if mediaField.storage.direct:
        attr = getattr(obj, field)
        filename = attr.get('name')
        if settings.STORAGE_LOCAL_X_ACCEL_REDIRECT:
            response = HttpResponse()
            response["Content-Disposition"] = (
                f"attachment; filename={filename}")
            response["Content-Type"] = attr.get('content-type', '')
            response['X-Accel-Redirect'] = (
                f"/filestore/{attr.get('key')}")
            return response
        response = FileResponse(
            open(url, 'rb'),
            as_attachment=True, filename=attr.get('name'))
        return response

    if get_bool(request.GET, 'redirect'):
        return redirect(url)
    return JsonResponse(dict(url=url))


class MediaRestViewMixin(ModelMixin):
    """Extension mixin to manage media fields."""

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.media_fields = set()
        for field in self.model._meta.fields:
            if isinstance(field, MediaField):
                self.media_fields.add(field)

    @rest_api(
        "/<int:pk>/<str:field>/upload-url/<str:filename>", method="get",
        priority_doc=-8)
    def field_media_upload_url(self, pk, field, filename, *args, **kwargs):
        """Entrypoint for
        :code:`GET /<int:pk>/<str:field>/upload-url/<str:filename>` route.

        Get the upload URL for a media file.

        :param pk: The object pk
        :param field: The field name
        :param filename: The file name
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get upload URL
            description: |
                Get upload URL for a `MediaField` field.

                You can upload a file ont the received URL. When the upload
                query is done, you have to confirm the process with an
                update request (`PUT`) on {model}. The body of this request
                must contain the name of the `MediaField` with the contents
                of the `file` value in the response of this request.

                1. `GET /<int:pk>/<str:field>/upload-url/<str:filename>`
                    → `data`
                2. `PUT data.url aFile Content-Type: aFile.type`
                3. `PUT /<int:pk> {{field: data.file}}`
            parameters extras:
                pk: The {model} pk.
                field: The {model} field name. Must be the name of
                    a `MediaField` field.
                filename: The desired filename.
            responses:
                200:
                    description: The upload URL
                    content:
                        application/json:
                            schema:
                                properties:
                                    url:
                                        type: string
                                        format: uri
                                    file:
                                        type: object
                                        properties:
                                            name:
                                                type: string
                                            key:
                                                type: string
        """
        obj = self.get_object(pk=pk)
        mediaField = get_medial_field(self.model, field)
        if mediaField.storage.direct:
            raise APIError("Unavailable for direct storage")
        try:
            res = mediaField.get_upload_url(self.request, obj, filename)
        except StorageException as e:  # pragma: no cover
            logger.exception(e)
            raise APIError(_("Unexpected storage error", status=500))
        return JsonResponse(res)

    @rest_api(
        "/<int:pk>/<str:field>", method="get",
        parameters=[parameters.MediaRedirect], priority_doc=-9)
    def field_media_get(self, pk, field, *args, **kwargs):
        """Entrypoint for :code:`GET /<int:pk>/<str:field>` route.

        Get the download URL for a media file.

        :param pk: The object pk
        :param field: The field name
        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get {model} file
            description: Get the URL for a media field file.
            parameters extras:
                pk: the {model} pk
                field: the {model} field name
            responses:
                200:
                    description: |
                        The file stream if storage is direct,
                        otherwise the file URL, only if `redirect` is `false`.
                    content:
                        application/json:
                            schema:
                                properties:
                                    url:
                                        type: string
                                        format: uri
                        application/octet-stream:
                            schema:
                                type: file
                302:
                    description: |
                        The redirect, for undirect storage
                        if `redirect` is `true`.
        """
        obj = self.get_object(pk=pk)
        return get_media_field_response(obj, field, self.request)

    def pre_save(self, obj, created=False):
        funcs = super().pre_save(obj, created=created)
        for field in self.media_fields:
            funcs.extend(field.media_pre_save(obj))
        return funcs


class MediaPermsRestViewMixin(MediaRestViewMixin):
    """Extension mixin to check permissions."""

    def field_media_upload_url_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('change'))

    def field_media_get_perm(self, *args, **kwargs):
        return self.request.user.has_perm(*self.get_model_perms('view'))



import inspect
import re

RE_CAMEL_CASE = s = re.compile(r"(_|-)+")


def camel_case(s):
    s = RE_CAMEL_CASE.sub(" ", s).title().replace(" ", "")
    return ''.join([s[0].lower(), s[1:]])


class ParameterGroup():
    parameters = []


class Parameter:
    _group = None

    @classmethod
    def id(cls):
        return cls.__name__

    @classmethod
    def as_parameter(cls, doc=None):
        res = dict(name=cls.name)
        res['in'] = cls.location
        description = doc or inspect.getdoc(cls)
        if description:
            res['description'] = description
        members = inspect.getmembers(
            cls, predicate=lambda x: not (
                inspect.ismethod(x)))
        for name, value in members:
            if name.startswith('_') or name == 'location':
                continue
            res[camel_case(name)] = value
        return res


class QueryParameter(Parameter):
    location = 'query'


def register_global_parameter(parameter):
    from . import __PARAMETERS__
    __PARAMETERS__.append(parameter)




import logging

from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

logger = logging.getLogger(__name__)


class UniqueConstraint(models.UniqueConstraint):
    def __init__(self, *args, message=None, **kwargs):
        self.message = message or _(
            "%(model_name)s with this %(field_labels)s already exists.")
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        path, args, kwargs = super().deconstruct()
        kwargs['message'] = self.message
        return path, args, kwargs


class JSONReprMixin():
    """A model mixin to manage JSON representation."""
    api = None
    url = None
    URL_AUTO_PK = True
    apidoc = {}

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._rel_data = {}
        self._after_save = []

    def get_url(self):
        return self.url

    def json_repr(self, **values):
        """Return the JSON representation of a mode instance.

        :param \\**values: Additional values to add
        :returns: The JSON representation
        :rtype: :class:`dict`
        """
        res = dict(
            pk=self.pk,
            resource_name=str(self))
        if hasattr(self, 'slug'):
            res['resource_slug'] = self.slug
        if hasattr(self, 'reference'):
            res['resource_reference'] = self.reference
        if self.api:
            res['resource_api'] = f"{self.api}/{self.pk}"
        url = self.get_url()
        if url:
            res['resource_url'] = (
                self.URL_AUTO_PK and f"{url}/{self.pk}" or url)
        res.update(**values)
        return res

    @classmethod
    def json_repr_schema(cls):
        """Return the JSON schema for the :code:`json_repr` method result.

        :returns: The JSON schema
        :rtype: :class:`dict`
        """
        from pfx.pfxcore.views.fields import FieldType
        res = dict(
            pk=dict(
                type=FieldType.to_apidoc(FieldType.from_model_field(
                    cls._meta.pk.__class__)),
                readonly=True),
            resource_name=dict(type='string', readonly=True))
        obj = cls()
        if hasattr(obj, 'slug'):
            res['resource_slug'] = dict(type='string', readonly=True)
        if hasattr(obj, 'reference'):
            res['resource_reference'] = dict(type='string', readonly=True)
        return res


class ErrorMessageMixin():
    """A model mixin to improve mode validation errors."""

    def validate_constraints(self, exclude=None):
        """Override validate_unique to allow to customize error message
        for :class:`UniqueConstraint`.
        """
        def objgetattr(obj, f):
            v = getattr(obj, f)
            if isinstance(v, models.Model) and v.pk:
                # Retrieve original instance to avoid having not stored
                # values used in instance repr().
                v = v.__class__._default_manager.get(pk=v.pk)
            return v

        try:
            return super().validate_constraints(exclude=exclude)
        except ValidationError as e:
            errors = e.update_error_dict({})
            non_field_errors = []
            for error in errors.get(NON_FIELD_ERRORS, []):
                if isinstance(error, ValidationError):
                    if error.code in ('unique', 'unique_together'):
                        model = error.params['model_class']
                        unique_check = error.params['unique_check']
                        unique = next(filter(
                            lambda c: isinstance(c, UniqueConstraint) and
                            c.fields == unique_check, model._meta.constraints
                        ), None)
                        if unique:
                            params = dict(error.params)
                            params.update({
                                f: objgetattr(self, f) for f in unique_check})
                            non_field_errors.append(ValidationError(
                                message=unique.message,
                                code=error.code,
                                params=params))
                            continue
                non_field_errors.append(error)
            errors[NON_FIELD_ERRORS] = non_field_errors
            raise ValidationError(errors)


class PFXModelMixin(JSONReprMixin, ErrorMessageMixin):
    pass



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.signals import pre_save
from django.utils.safestring import mark_safe

import nh3


class RichTextField(models.TextField):
    """
    A TextField that automatically sanitizes HTML input using nh3.

    Args:
        nh3_config (str|None): Key of the NH3_CONFIGS dict from settings.
            Set to None to disable sanitization for this field.
    """
    def __init__(self, *args, nh3_config='default', **kwargs):
        super().__init__(*args, **kwargs)
        if nh3_config is None:
            self.nh3_kwargs = None
        else:
            if (not hasattr(settings, 'NH3_CONFIGS') or
                    nh3_config not in settings.NH3_CONFIGS):
                raise ImproperlyConfigured(
                    f"Missing NH3_CONFIGS['{nh3_config}'] configuration. "
                    "You must use an existing configuration or set "
                    "nh3_config=None to explicitly disable sanitization "
                    "for this field.")
            self.nh3_kwargs = settings.NH3_CONFIGS[nh3_config]

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        # Values are sanitized before saving, so any value returned from the DB
        # is safe to render unescaped.
        return mark_safe(value)

    def contribute_to_class(self, cls, name, private_only=False):
        super().contribute_to_class(cls, name, private_only=private_only)

        if self.nh3_kwargs:
            # connect a pre_save signal for this field
            def sanitize(sender, instance, **kwargs):
                value = getattr(instance, name)
                if value:
                    clean_value = nh3.clean(value, **self.nh3_kwargs)
                    setattr(instance, name, mark_safe(clean_value))

            pre_save.connect(sanitize, sender=cls, weak=False)



from datetime import date
from pathlib import Path
from unittest.mock import MagicMock, patch

from django.db import connection
from django.test import TestCase, override_settings

from pfx.pfxcore.test import APIClient, MockBoto3Client, TestAssertMixin
from tests.models import Author, Book, BookType


@override_settings(
    STORAGE_LOCAL_ROOT='/tmp/django-pfx-filestore',
    STORAGE_LOCAL_X_ACCEL_REDIRECT=False)
class BasicAPITest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')
        with connection.cursor() as cursor:
            cursor.execute("create extension if not exists unaccent;")

    @classmethod
    def setUpTestData(cls):
        cls.type_sf = BookType.objects.create(
            name="Science fiction", slug="sf")
        cls.type_fantastique = BookType.objects.create(
            name="Fantastique", slug="fantastique")
        cls.type_fantasy = BookType.objects.create(
            name="Fantastique", slug="fantastique")
        cls.type_romance = BookType.objects.create(
            name="Romance", slug="romance")

        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        cls.author1_book1 = Book.objects.create(
            author=cls.author1,
            name="The Hobbit",
            pub_date=date(1937, 1, 1))
        cls.author1_book2 = Book.objects.create(
            author=cls.author1,
            name="The Fellowship of the Ring",
            pub_date=date(1954, 1, 1))
        cls.author1_book3 = Book.objects.create(
            author=cls.author1,
            name="The Two Towers",
            pub_date=date(1954, 1, 1))
        cls.author1_book4 = Book.objects.create(
            author=cls.author1,
            name="The Return of the King",
            pub_date=date(1955, 1, 1))

        cls.author2 = Author.objects.create(
            first_name='Philip Kindred',
            last_name='Dick',
            science_fiction=True,
            slug='philip-k-dick')
        cls.author2.types.set([cls.type_sf, cls.type_fantastique])

        cls.author3 = Author.objects.create(
            first_name='Isaac',
            last_name='Asimov',
            science_fiction=True,
            slug='isaac-asimov')
        cls.author3.types.set([cls.type_sf])
        cls.author3_book1 = Book.objects.create(
            author=cls.author3,
            name="The Caves of Steel",
            pub_date=date(1954, 1, 1),
            pages=224,
            rating=4.6)
        cls.author3_book2 = Book.objects.create(
            author=cls.author3,
            name="The Naked Sun",
            pages=304,
            pub_date=date(1957, 1, 1))
        cls.author3_book3 = Book.objects.create(
            author=cls.author3,
            name="The Robots of Dawn",
            pub_date=date(1983, 1, 1))

        cls.author4 = Author.objects.create(
            first_name='Joanne',
            last_name='Rowling',
            science_fiction=False,
            gender='female',
            slug='j-k-rowling')

    def test_get_list(self):
        response = self.client.get('/api/authors?items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_get_list_extra_meta(self):
        response = self.client.get('/api/authors-extra-meta?items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)
        self.assertJE(response, 'meta.test', "Hello world")

    @override_settings(PFX_MAX_LIST_RESULT_SIZE=10)
    def test_get_list_global_limit(self):
        for i in range(10):
            Book.objects.create(
                author=self.author1,
                name=f"Book {i}",
                pub_date=date(1900, 1, 1))

        response = self.client.get('/api/books?items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 17)
        self.assertSize(response, 'items', 10)

        response = self.client.get(
            '/api/books?subset=pagination&page-size=50&items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 17)
        self.assertJE(response, 'meta.subset.page_size', 10)
        self.assertSize(response, 'items', 10)

        response = self.client.get(
            '/api/books?subset=offset&page-size=50&items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 17)
        self.assertJE(response, 'meta.subset.limit', 10)
        self.assertSize(response, 'items', 10)

    def test_get_list_order_asc(self):
        response = self.client.get('/api/authors?order=last_name')
        self.assertRC(response, 200)

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Asimov', 'Dick', 'Rowling', 'Tolkien'])

    def test_get_list_order_desc(self):
        response = self.client.get('/api/authors?order=-last_name')
        self.assertRC(response, 200)

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Tolkien', 'Rowling', 'Dick', 'Asimov'])

    def test_get_list_order_multi(self):
        response = self.client.get('/api/authors?order=gender,last_name')
        self.assertRC(response, 200)

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Rowling', 'Asimov', 'Dick', 'Tolkien'])

    def test_get_list_order_multi_desc(self):
        response = self.client.get('/api/authors?order=gender,-last_name')
        self.assertRC(response, 200)

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Rowling', 'Tolkien', 'Dick', 'Asimov'])

    def test_get_list_order_invalid(self):
        response = self.client.get('/api/authors?order=invalid_field')
        self.assertRC(response, 400)

        self.assertIn(
            "Cannot resolve keyword 'invalid_field' into field.",
            self.get_val(response, 'message'))

    def test_get_list_order_with_order_mapping(self):
        response = self.client.get(
            f'/api/books?author={self.author3.pk}&order=pages')
        self.assertRC(response, 200)
        pages = [
            self.get_val(item, 'pages')
            for item in self.get_val(response, 'items')]
        self.assertEqual(pages, [None, 224, 304])

        response = self.client.get(
            f'/api/books?author={self.author3.pk}&order=%2Bpages')
        self.assertRC(response, 200)
        pages = [
            self.get_val(item, 'pages')
            for item in self.get_val(response, 'items')]
        self.assertEqual(pages, [None, 224, 304])

        response = self.client.get(
            f'/api/books?author={self.author3.pk}&order=-pages')
        self.assertRC(response, 200)
        pages = [
            self.get_val(item, 'pages')
            for item in self.get_val(response, 'items')]
        self.assertEqual(pages, [304, 224, None])

    def test_search_list(self):
        response = self.client.get('/api/authors?search=isaac&items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.first_name', 'Isaac')
        self.assertJE(response, 'items.@0.last_name', 'Asimov')
        self.assertJE(response, 'items.@0.gender.value', 'male')
        self.assertJE(response, 'items.@0.gender.label', 'Male')

        response = self.client.get(
            '/api/authors?search=isaac&search=tolkien&items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 2)

    def test_meta_list(self):
        response = self.client.get('/api/authors/meta/list')
        self.assertRC(response, 200)
        self.assertJE(response, 'model.app', 'tests')
        self.assertJE(response, 'model.name', 'author')
        self.assertJE(response, 'model.object', 'Author')

        self.assertJE(response, 'filters.@0.name', 'book_type')
        self.assertJE(response, 'filters.@0.items', [
            {
                'empty_value': False,
                "is_group": False,
                "label": "Science Fiction",
                "name": "science_fiction",
                "technical": True,
                "type": "BooleanField"
            },
            {
                'empty_value': True,
                "is_group": False,
                "label": "Heroic Fantasy",
                "name": "heroic_fantasy",
                "technical": False,
                "type": "BooleanField"
            },
            {
                'empty_value': False,
                "is_group": False,
                "label": "Types",
                "name": "types",
                "related_model": "BookType",
                "api": "/book-types",
                "technical": False,
                "type": "ModelObject"
            }
        ])
        self.assertJE(response, 'filters.@1.name', 'custom')
        self.assertJE(response, 'filters.@1.items', [
            {
                'empty_value': False,
                "is_group": False,
                "label": "Last Name",
                "name": "last_name",
                "technical": False,
                "type": "BooleanField"
            },
            {
                'empty_value': False,
                "is_group": False,
                "label": "First Name",
                "name": "first_name",
                "technical": False,
                "type": "CharField"
            },
            {
                "choices": [
                    {
                        "label": "Male",
                        "value": "male"
                    },
                    {
                        "label": "Female",
                        "value": "female"
                    }
                ],
                'empty_value': False,
                "is_group": False,
                "label": "Gender",
                "name": "gender",
                "technical": False,
                "type": "CharField"
            },
            {
                "choices": [
                    {
                        "label": "Tolkien",
                        "value": "Tolkien"
                    },
                    {
                        "label": "Asimov",
                        "value": "Asimov"
                    }
                ],
                'empty_value': False,
                "is_group": False,
                "label": "Tolkien or Asimov",
                "name": "last_name_choices",
                "technical": False,
                "type": "CharField"
            }
        ])

        response = self.client.get('/api/books/meta/list?filters=1&orders=1')
        self.assertRC(response, 200)

        self.assertJE(response, 'filters.@0.items.@0', {
            "empty_value": False,
            "is_group": False,
            "label": "Author",
            "name": "author",
            "related_model": "Author",
            "api": "/authors",
            "technical": False,
            "type": "ModelObject"
        })

        self.assertJIn(response, 'orders', 'pk')
        self.assertJIn(response, 'orders', 'name')
        self.assertJIn(response, 'orders', 'author')
        self.assertJIn(response, 'orders', 'author__pk')
        self.assertJIn(response, 'orders', 'author__first_name')

    def test_filter_func_bool(self):
        response = self.client.get(
            '/api/authors?heroic_fantasy=true&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.resource_name',
                                'John Ronald Reuel Tolkien')

        response = self.client.get(
            '/api/authors?heroic_fantasy=false&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

        response = self.client.get(
            '/api/authors?heroic_fantasy=false&heroic_fantasy=true'
            '&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_filter_func_date(self):
        response = self.client.get(
            '/api/books?pub_date_gte=1955-01-01&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

    def test_filter_func_char_choices(self):
        response = self.client.get(
            '/api/authors?last_name_choices=Tolkien&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)

    def test_model_filter_bool(self):
        response = self.client.get(
            '/api/authors?science_fiction=true&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 2)

        response = self.client.get(
            '/api/authors?science_fiction=invalid&items=1&count=1')
        self.assertRC(response, 400)
        self.assertJE(
            response, 'message', "Invalid value for Science Fiction filter")

    def test_model_filter_char(self):
        response = self.client.get(
            '/api/authors?first_name=Isaac&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.first_name', "Isaac")

    def test_model_filter_integer(self):
        response = self.client.get('/api/books?pages=224&items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.name', "The Caves of Steel")

        response = self.client.get('/api/books?pages=invalid&items=1&count=1')
        self.assertRC(response, 400)
        self.assertJE(response, 'message', "Invalid value for Pages filter")

    def test_model_filter_float(self):
        response = self.client.get('/api/books?rating=4.6&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.name', "The Caves of Steel")

        response = self.client.get('/api/books?rating=invalid&items=1&count=1')
        self.assertRC(response, 400)
        self.assertJE(response, 'message', "Invalid value for Rating filter")

    def test_model_filter_date(self):
        response = self.client.get(
            '/api/books?pub_date=1954-01-01&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

        response = self.client.get(
            '/api/books?pub_date=invalid&items=1&count=1')
        self.assertRC(response, 400)
        self.assertJE(response, 'message', "Invalid value for Pub Date filter")

    def test_model_filter_manytomany(self):
        response = self.client.get(
            f'/api/authors?types={self.type_romance.pk}&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 0)

        response = self.client.get(
            f'/api/authors?types={self.type_fantastique.pk}&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.pk', self.author2.pk)

        response = self.client.get(
            f'/api/authors?types={self.type_sf.pk}&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 2)
        self.assertJE(response, 'items.@0.pk', self.author3.pk)
        self.assertJE(response, 'items.@1.pk', self.author2.pk)

        # Check multiple values filtered with OR condition.
        response = self.client.get(
            f'/api/authors?types={self.type_fantastique.pk}'
            f'&types={self.type_romance.pk}&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.pk', self.author2.pk)

        response = self.client.get(
            f'/api/authors?types={self.type_fantastique.pk}'
            f'&types={self.type_sf.pk}&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 2)
        self.assertJE(response, 'items.@0.pk', self.author3.pk)
        self.assertJE(response, 'items.@1.pk', self.author2.pk)

        response = self.client.get(
            '/api/authors?types=invalid&items=1&count=1')
        self.assertRC(response, 400)
        self.assertJE(response, 'message', "Invalid value for Types filter")

    def test_model_filter_decimal(self):
        # TODO: Wait Decimal implementation
        pass

    def test_model_filter_foreign_key(self):
        response = self.client.get(
            f'/api/books?author={self.author3.pk}&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        for item in response.json_content['items']:
            self.assertEqual(item['author']['pk'], self.author3.pk)

        response = self.client.get(
            '/api/books?author=invalid&items=1&count=1')
        self.assertRC(response, 400)
        self.assertJE(response, 'message', "Invalid value for Author filter")

    def test_model_filter_foreign_key_null(self):
        book_type = BookType.objects.create(
            name="Epic Fantasy",
            slug="epic-fantasy")
        self.author1_book1.type = book_type
        self.author1_book1.save()

        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type={book_type.pk}'
            '&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)

        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type=null&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        for item in response.json_content['items']:
            self.assertEqual(item['type'], None)
        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type=0&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 0)
        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type=&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        for item in response.json_content['items']:
            self.assertEqual(item['type'], None)

        response = self.client.get(
            f'/api/books?author={self.author1.pk}&'
            f'type={book_type.pk}&type=null&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)
        for item in response.json_content['items']:
            self.assertTrue(
                item['type'] is None or item['type']['pk'] == book_type.pk)

    def test_model_filter_foreign_key_text_search(self):
        response = self.client.get(
            '/api/books?author*=asimov&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

        # Test multiple values
        response = self.client.get(
            '/api/books?author*=asimov&author*=tolkien&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 7)

        # Test empty search
        response = self.client.get(
            '/api/books?author*=&items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 7)

    def test_model_filter_char_choices(self):
        response = self.client.get('/api/authors?gender=male&items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

    def test_model_filter_func_bool(self):
        response = self.client.get(
            '/api/authors?last_name=true&items=1&count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_get_list_without_pagination(self):
        response = self.client.get('/api/authors?count=1')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_get_list_with_subset(self):
        for i in range(1, 11):
            Author.objects.create(
                first_name='Vladimir',
                last_name=f'Ottor {i}',
                science_fiction=True,
                slug=f'vald-ottor-{i}')

        response = self.client.get('/api/authors?subset=offset&limit=5')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.subset.offset', 0)
        self.assertJE(response, 'meta.subset.count', 14)
        item4_pk = response.json_content['items'][3]['pk']

        response = self.client.get(
            '/api/authors?subset=offset&offset=3&limit=5')
        self.assertRC(response, 200)
        # With offset 3 the first item must be the same as the 4th
        # in the request with offest 0.
        self.assertJE(response, 'items.@0.pk', item4_pk)
        self.assertJE(response, 'meta.subset.count', 14)

    def test_select_list(self):
        response = self.client.get(
            '/api/authors?mode=select&items=true&limit=10')
        self.assertRC(response, 200)
        self.assertSize(response, 'items', 4)

    def test_select_list_search(self):
        response = self.client.get(
            '/api/authors?mode=select&subset=offset&limit=10&search=Isaac')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.subset.count', 1)
        self.assertJE(response, 'items.@0.resource_name', "Isaac Asimov")

    def test_select_list_max(self):
        response = self.client.get(
            '/api/authors?mode=select&subset=offset&items=true&limit=3')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.subset.count', 4)
        self.assertSize(response, 'items', 3)

    def test_meta_service(self):
        response = self.client.get('/api/authors/meta')
        self.assertRC(response, 200)
        self.assertJE(response, 'model.app', 'tests')
        self.assertJE(response, 'model.name', 'author')
        self.assertJE(response, 'model.object', 'Author')
        self.assertJE(response, 'fields.first_name.type', 'CharField')
        self.assertJE(response, 'fields.first_name.name', 'First Name')
        self.assertJE(response, 'fields.last_name.type', 'CharField')
        self.assertJE(response, 'fields.last_name.name', 'Last Name')
        self.assertJE(response, 'fields.books.type', 'ModelObjectList')
        self.assertJE(response, 'fields.books.name', 'Books')
        self.assertJE(response, 'fields.created_at.type', 'DateField')
        self.assertJE(response, 'fields.created_at.name', 'Created at')
        self.assertJE(response, 'fields.name_length.type', 'IntegerField')
        self.assertJE(response, 'fields.name_length.name', 'Name Length')
        self.assertJE(response, 'fields.types.type', 'ModelObjectList')
        self.assertJE(response, 'fields.types.name', 'Types')
        response = self.client.get('/api/books/meta')
        self.assertJE(response, 'fields.author.type', 'ModelObject')
        self.assertJE(response, 'fields.author.model', 'tests.Author')
        self.assertJE(response, 'fields.author.api', '/authors')
        self.assertJE(
            response, 'fields.read_time.type', 'MinutesDurationField')
        self.assertJE(response, 'fields.read_time.name', 'Read Time')

    def test_meta_service_fr(self):
        response = self.client.get('/api/authors/meta', locale='fr')
        self.assertRC(response, 200)
        self.assertJE(response, 'fields.first_name.type', 'CharField')
        self.assertJE(response, 'fields.first_name.name', 'Prénom')
        self.assertJE(response, 'fields.last_name.type', 'CharField')
        self.assertJE(response, 'fields.last_name.name', 'Nom')
        response = self.client.get('/api/books/meta')
        self.assertJE(
            response, 'fields.read_time.type', 'MinutesDurationField')
        self.assertJE(response, 'fields.read_time.name', 'Read Time')

    def test_get_detail_by_id(self):
        response = self.client.get(f'/api/authors/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', self.author1.first_name)
        self.assertJE(response, 'last_name', self.author1.last_name)
        self.assertJE(response, 'name_length', 25)
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')

    def test_get_detail_by_slug(self):
        response = self.client.get(f'/api/authors/slug/{self.author1.slug}')
        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', self.author1.first_name)
        self.assertJE(response, 'last_name', self.author1.last_name)
        self.assertJE(response, 'name_length', 25)
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')

    def test_get_detail_navigation_meta(self):
        response = self.client.get(
            f'/api/authors/{self.author3.pk}?navigation=1&order=last_name')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.previous_pk', self.author1.pk)
        self.assertJE(response, 'meta.next_pk', self.author2.pk)
        self.assertJE(response, 'meta.index', 1)
        self.assertJE(response, 'meta.count', 4)

        response = self.client.get(
            f'/api/authors/{self.author2.pk}?navigation=1&order=last_name')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.previous_pk', self.author3.pk)
        self.assertJE(response, 'meta.next_pk', self.author4.pk)
        self.assertJE(response, 'meta.index', 2)
        self.assertJE(response, 'meta.count', 4)

        response = self.client.get(
            f'/api/authors/{self.author2.pk}?navigation=1&order=-last_name')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.previous_pk', self.author4.pk)
        self.assertJE(response, 'meta.next_pk', self.author3.pk)
        self.assertJE(response, 'meta.index', 3)
        self.assertJE(response, 'meta.count', 4)

    def test_get_date_format(self):
        response = self.client.get(f'/api/books/{self.author1_book1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'pub_date', '1937-01-01')

    def test_get_detail_with_rel_object(self):
        response = self.client.get(f'/api/books/{self.author3_book1.pk}')

        self.assertRC(response, 200)
        self.assertJE(response, 'author.pk', self.author3.pk)
        self.assertJE(response, 'author.resource_name', "Isaac Asimov")
        self.assertJE(response, 'author.resource_slug', "isaac-asimov")
        self.assertJEExists(response, 'author.resource_slug')
        self.assertJENotExists(response, 'author.hello')
        self.assertJENotExists(response, 'author.last_name')

    def test_get_detail_with_rel_object_custom(self):
        response = self.client.get(
            f'/api/books-custom-author/{self.author3_book1.pk}')

        self.assertRC(response, 200)
        self.assertJE(response, 'author.pk', self.author3.pk)
        self.assertJE(response, 'author.resource_name', "Isaac Asimov")
        self.assertJE(response, 'author.resource_slug', "isaac-asimov")
        self.assertJE(response, 'author.hello', "World")
        self.assertJE(response, 'author.last_name', "Asimov")

    def test_get_detail_manytomany(self):
        response = self.client.get(f'/api/authors/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertSize(response, 'types', 0)

        response = self.client.get(f'/api/authors/{self.author2.pk}')
        self.assertRC(response, 200)
        self.assertSize(response, 'types', 2)
        self.assertJE(response, 'types.@0.pk', self.type_sf.pk)
        self.assertJE(response, 'types.@1.pk', self.type_fantastique.pk)

        response = self.client.get(f'/api/authors/{self.author3.pk}')
        self.assertRC(response, 200)
        self.assertSize(response, 'types', 1)
        self.assertJE(response, 'types.@0.pk', self.type_sf.pk)

    def test_get_non_existant_record(self):
        response = self.client.get('/api/authors/999999')
        self.assertRC(response, 404)

    def test_get_non_existant_record_by_slug(self):
        response = self.client.get('/api/authors/slug/not-existent')
        self.assertRC(response, 404)

    def test_create(self):
        response = self.client.post(
            '/api/authors', dict(
                first_name='Arthur Charles',
                last_name='Clarke',
                name_length=1,
                gender='male',
                create_comment='CREATE COMMENT',
                update_comment='UPDATE COMMENT',
                website=None,
                slug='arthur-c-clarke',
                types=[self.type_sf.pk, dict(pk=self.type_fantastique.pk)]))

        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', 'Arthur Charles')
        self.assertJE(response, 'last_name', 'Clarke')
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')
        self.assertJE(response, 'create_comment', 'CREATE COMMENT')
        self.assertNJE(response, 'update_comment', 'UPDATE COMMENT')
        self.assertJE(response, 'website', '')
        self.assertSize(response, 'types', 2)
        self.assertJE(response, 'types.@0.pk', self.type_sf.pk)
        self.assertJE(response, 'types.@1.pk', self.type_fantastique.pk)

    def test_create_file(self):
        response = self.client.post('/api/books', dict(
            name="Test Book",
            author=self.author1.pk,
            pub_date='2000-01-01',
            local_file=dict(
                name='test.txt',
                base64='data:text/plain;charset=utf-8;'
                'base64,VGVzdCBjb250ZW50')))
        book_pk = self.get_val(response, 'pk')
        self.assertRC(response, 200)
        self.assertJE(response, 'local_file.content-length', '12')
        self.assertJE(response, 'local_file.content-type', 'text/plain')
        self.assertJE(
            response, 'local_file.key',
            f"Book/{book_pk}/bca20547e94049e1ffea27223581c567022a5774.txt")
        self.assertJE(response, 'local_file.name', 'test.txt')
        self.assertJE(
            response, 'local_file.url', f"/books/{book_pk}/local_file")
        self.assertTrue(Path(
            f'/tmp/django-pfx-filestore/Book/{book_pk}/'
            'bca20547e94049e1ffea27223581c567022a5774.txt').exists())

        # Replace file
        response = self.client.put(f'/api/books/{book_pk}', dict(
            local_file=dict(
                name='test2.txt',
                base64='data:text/plain;charset=utf-8;'
                'base64,TmV3IGNvbnRlbnQ=')))
        self.assertRC(response, 200)
        self.assertJE(
            response, 'local_file.key',
            f"Book/{book_pk}/af9f06b2b3b1546ac44f4a02994d0ef09e074b91.txt")
        self.assertJE(response, 'local_file.name', 'test2.txt')
        self.assertFalse(Path(
            f'/tmp/django-pfx-filestore/Book/{book_pk}/'
            'bca20547e94049e1ffea27223581c567022a5774.txt').exists())
        self.assertTrue(Path(
            f'/tmp/django-pfx-filestore/Book/{book_pk}/'
            'af9f06b2b3b1546ac44f4a02994d0ef09e074b91.txt').exists())

        # Delete file
        response = self.client.put(f'/api/books/{book_pk}', dict(
            local_file=None))
        self.assertRC(response, 200)
        self.assertJE(response, 'local_file', None)
        self.assertFalse(Path(
            f'/tmp/django-pfx-filestore/Book/{book_pk}/'
            'af9f06b2b3b1546ac44f4a02994d0ef09e074b91.txt').exists())

    def test_create_null_values(self):
        response = self.client.post(
            '/api/books', dict(
                name="Test Book",
                author=self.author1.pk,
                type=None,
                pub_date='2000-01-01',
                pages=None,
                rating=None,
                reference=None,
                read_time=None))
        self.assertRC(response, 200)

    def test_create_enum(self):
        response = self.client.post(
            '/api/authors', {
                'first_name': 'Arthur Charles',
                'last_name': 'Clarke',
                'name_length': 1,
                'gender': {'value': 'male'},
                'slug': 'arthur-c-clarke'})

        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', 'Arthur Charles')
        self.assertJE(response, 'last_name', 'Clarke')
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')

    def test_create_enum_bad_value(self):
        response = self.client.post(
            '/api/authors', {
                'first_name': 'Arthur Charles',
                'last_name': 'Clarke',
                'name_length': 1,
                'gender': {'value': 'writer'},
                'slug': 'arthur-c-clarke'})

        self.assertRC(response, 422)
        self.assertJE(
            response, 'gender.@0', "Value 'writer' is not a valid choice.")

    def test_create_unique_custom_message(self):
        response = self.client.post(
            '/api/books', dict(
                name="The Hobbit",
                author=self.author1.pk))
        self.assertRC(response, 422)
        self.assertJE(
            response, '__all__.@0',
            'The Hobbit already exists for John Ronald Reuel Tolkien')

    def test_create_invalid_date(self):
        response = self.client.post(
            '/api/books', dict(
                name="The Silmarillion",
                author=self.author1.pk,
                pub_date="19777-09-15"))
        self.assertRC(response, 422)
        self.assertJE(
            response, 'pub_date.@0',
            '“19777-09-15” value has an invalid date format. '
            'It must be in YYYY-MM-DD format.')

    def test_update(self):
        response = self.client.put(
            f'/api/authors/{self.author1.pk}',
            {'pk': self.author2.pk,  # pk and id must be ignored
             'created_at': '2021-01-01',  # created_at must be ignored because
                                          # it is a readonly field.
             'first_name': 'J. R. R.',
             'name_length': 1,
             'gender': 'female',
             'create_comment': 'CREATE COMMENT UPDATED',
             'update_comment': 'UPDATE COMMENT UPDATED',
             'slug': 'j-r-r-tolkien'})  # slug must be updated})

        self.assertRC(response, 200)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.first_name, 'J. R. R.')
        self.assertEqual(self.author1.last_name, 'Tolkien')
        self.assertEqual(self.author1.slug, 'j-r-r-tolkien')
        self.assertEqual(self.author1.gender, 'female')
        self.assertNotEqual(self.author1.created_at, '2021-01-01 11:30:00')
        self.assertNotEqual(
            self.author1.create_comment, 'CREATE COMMENT UPDATED')
        self.assertEqual(
            self.author1.update_comment, 'UPDATE COMMENT UPDATED')

        response = self.client.put(
            f'/api/authors/{self.author1.pk}', {'gender': {'value': 'male'}})
        self.assertRC(response, 200)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.gender, 'male')

    def test_update_manytomny(self):
        self.assertEqual(self.author1.types.count(), 0)

        response = self.client.put(
            f'/api/authors/{self.author1.pk}', dict(
                types=[self.type_fantasy.pk]))
        self.assertRC(response, 200)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.types.count(), 1)
        self.assertEqual(self.author1.types.all()[0].pk, self.type_fantasy.pk)

        response = self.client.put(
            f'/api/authors/{self.author1.pk}', dict(types=[]))
        self.assertRC(response, 200)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.types.count(), 0)

    def test_update_enum_bad_value(self):
        response = self.client.put(
            f'/api/authors/{self.author1.pk}', {
                'gender': {'value': 'writer'}})

        self.assertRC(response, 422)
        self.assertJE(
            response, 'gender.@0', "Value 'writer' is not a valid choice.")

    def test_empty_number_fields(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring, Deluxe Edition',
                'author': self.author1.pk,
                'pub_date': '1954-07-29',
                'pages': '',
                'rating': '',
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'pages', None)
        self.assertJE(response, 'rating', None)

        response = self.client.post(
            '/api/books', {
                'name': 'The Two Towers, Deluxe Edition',
                'author': self.author1.pk,
                'pub_date': '1954-07-29',
                'pages': 500,
                'rating': 5.35,
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'pages', 500)
        self.assertJE(response, 'rating', 5.35)

        response = self.client.put(
            f"/api/books/{response.json_content['pk']}", {
                'pages': '',
                'rating': '',
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'pages', None)
        self.assertJE(response, 'rating', None)

        response = self.client.put(
            f"/api/books/{response.json_content['pk']}", {
                'pages': 600,
                'rating': 6.34,
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'pages', 600)
        self.assertJE(response, 'rating', 6.34)

    def test_delete(self):
        response = self.client.delete(
            f'/api/authors/{self.author2.pk}')

        self.assertRC(response, 200)

        author = Author.objects.filter(pk=self.author2.pk)
        self.assertEqual(author.count(), 0)

    def test_delete_with_wrong_key(self):
        response = self.client.delete(
            '/api/authors/99999')
        self.assertRC(response, 404)

    def test_create_with_foreignkey(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring, Deluxe Edition',
                'author': self.author1.pk,
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })
        self.assertRC(response, 200)
        self.assertJE(
            response, 'name', 'The Fellowship of the Ring, Deluxe Edition')
        self.assertJE(response, 'author.pk', self.author1.pk)
        self.assertJE(response, 'pub_date', '1954-07-29')
        self.assertNJE(response, 'created_at', '1954-07-29')

    def test_create_with_foreignkey_resource(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring, Deluxe Edition',
                'author': {
                    'pk': self.author1.pk,
                    'resource_name': "Author One"},
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'author.pk', self.author1.pk)

    def test_create_with_wrong_foreignkey(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring, Deluxe Edition',
                'author': 999999,
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })
        self.assertRC(response, 422)
        self.assertJE(response, 'author', ['Author not found.'])

    def test_update_with_foreignkey(self):
        response = self.client.put(
            f'/api/books/{self.author1_book2.pk}', {
                'name': 'The Two Towers, Deluxe Edition',
                'pub_date': '1954-11-11',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'The Two Towers, Deluxe Edition')
        self.assertJE(response, 'author.pk', self.author1.pk)
        self.assertJE(response, 'pub_date', '1954-11-11')

        response = self.client.put(
            f'/api/books/{self.author1_book2.pk}', {
                'name': 'The Man in the High Castle',
                'author': self.author2.pk,
                'author_id': self.author3.pk,  # must be ignored
                'pub_date': '1962-10-01',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'The Man in the High Castle')
        self.assertJE(response, 'author.pk', self.author2.pk)
        self.assertJE(response, 'pub_date', '1962-10-01')

        response = self.client.put(
            f'/api/books/{self.author1_book2.pk}', {
                'name': 'A Scanner Darkly',
                'author': {
                    'pk': self.author2.pk,
                    'resource_name': 'Philip Kindred Dick'},
                'pub_date': '1977-01-01',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'A Scanner Darkly')
        self.assertJE(response, 'author.pk', self.author2.pk)
        self.assertJE(response, 'pub_date', '1977-01-01')

    def test_update_with_wrong_key_and_foreignkey(self):
        response = self.client.put(
            '/api/books/99999', {
                'name': 'The Two Towers',
                'pub_date': '1954-11-11'})
        self.assertRC(response, 404)

        response = self.client.put(
            f'/api/books/{self.author1_book2.pk}', {
                'name': 'The Two Towers',
                'author': 999999,
                'pub_date': '1954-11-11'})
        self.assertRC(response, 422)
        self.assertJE(response, 'author', ['Author not found.'])

    def test_delete_with_foreignkey(self):
        response = self.client.delete(
            f'/api/authors/{self.author1.pk}')

        self.assertRC(response, 400)
        self.author1_book2.refresh_from_db()

    def test_create_validation(self):
        response = self.client.post(
            '/api/books', {
                'name': '',
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })

        self.assertRC(response, 422)
        self.assertJE(response, 'name',
                      ['This field cannot be blank.'])
        self.assertJE(response, 'author',
                      ['This field cannot be null.'])

    def test_update_validation(self):
        response = self.client.put(
            f'/api/books/{self.author1_book2.pk}', {
                'name': '',
                'pub_date': '1954-11-11',
            })

        self.assertRC(response, 422)
        self.assertJE(response, 'name',
                      ['This field cannot be blank.'])

    def test_create_related_field(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring, Deluxe Edition',
                'author': self.author1.pk,
                'author__last_name': "Teulkien",
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })
        self.assertRC(response, 200)
        self.assertJE(
            response, 'name', 'The Fellowship of the Ring, Deluxe Edition')
        self.assertJE(response, 'author.pk', self.author1.pk)
        self.assertJE(response, 'pub_date', '1954-07-29')
        self.assertNJE(response, 'created_at', '1954-07-29')
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.last_name, "Tolkien")

    def test_update_related_field(self):
        response = self.client.put(
            f'/api/books/{self.author1_book2.pk}', {
                'name': 'The Two Towers, Deluxe Edition',
                'author__last_name': "Teulkien",
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'The Two Towers, Deluxe Edition')
        self.assertJE(response, 'author.pk', self.author1.pk)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.last_name, "Tolkien")

    def test_custom_repr(self):
        book_type = BookType.objects.create(
            name="Epic Fantasy",
            slug="epic-fantasy",
        )
        self.author1_book2.type = book_type
        self.author1_book2.save()

        response = self.client.get(f'/api/books/{self.author1_book2.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'type.resource_name', book_type.name)
        self.assertJE(response, 'type.resource_slug', book_type.slug)

        response = self.client.get(f'/api/book-types/{book_type.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'resource_name', book_type.name)
        self.assertJE(response, 'resource_slug', book_type.slug)
        self.assertJE(response, 'name', book_type.name)
        self.assertJE(response, 'slug', book_type.slug)

    @override_settings(
        STORAGE_S3_AWS_REGION="fake-region",
        STORAGE_S3_AWS_ACCESS_KEY="FAKE",
        STORAGE_S3_AWS_SECRET_KEY="FAKE-SECRET",
        STORAGE_S3_AWS_S3_BUCKET="dragonfly.fake",
        STORAGE_S3_AWS_GET_URL_EXPIRE=300,
        STORAGE_S3_AWS_PUT_URL_EXPIRE=300)
    @patch("boto3.client", MagicMock(return_value=MockBoto3Client()))
    def test_media_field(self):
        response = self.client.get(
            f'/api/books/{self.author1_book1.pk}/cover/upload-url/'
            'cover.png?content-type=image/png')
        self.assertRC(response, 200)
        self.assertJE(response, 'file.key', "Book/1/cover.png")
        self.assertJE(response, 'file.name', "cover.png")
        self.assertJE(
            response, 'url', "http://dragonfly.fake/Book/1/cover.png")

        response = self.client.put(f'/api/books/{self.author1_book1.pk}', dict(
            cover=response.json_content['file']))
        self.assertRC(response, 200)
        self.assertJE(response, 'cover.key', "Book/1/cover.png")
        self.assertJE(response, 'cover.name', "cover.png")
        self.assertJE(response, 'cover.content-length', 1000)
        self.assertJE(response, 'cover.content-type', "image/png")

        response = self.client.get(
            f'/api/books/{self.author1_book1.pk}/cover')
        self.assertRC(response, 200)
        self.assertJE(
            response, 'url', "http://dragonfly.fake/Book/1/cover.png")

        response = self.client.get(
            f'/api/books/{self.author1_book1.pk}/cover?redirect=false')
        self.assertRC(response, 200)
        self.assertJE(
            response, 'url', "http://dragonfly.fake/Book/1/cover.png")

        response = self.client.get(
            f'/api/books/{self.author1_book1.pk}/cover?redirect=true')
        self.assertRedirects(
            response, "http://dragonfly.fake/Book/1/cover.png",
            fetch_redirect_response=False)

        with patch.object(
                MockBoto3Client, 'delete_object',
                return_value=None) as mock_delete:
            response = self.client.delete(
                f'/api/books/{self.author1_book1.pk}')
            mock_delete.assert_called_with(
                Bucket='dragonfly.fake',
                Key=f'Book/{self.author1_book1.pk}/cover.png')

    def test_annotate_meta_service(self):
        response = self.client.get('/api/authors-annotate/meta')
        self.assertRC(response, 200)
        self.assertJE(response, 'fields.first_name.type', 'CharField')
        self.assertJE(response, 'fields.first_name.name', 'First Name')
        self.assertJE(response, 'fields.books_count.type', None)
        self.assertJE(response, 'fields.books_count.name', 'books_count')
        self.assertJE(response, 'fields.books_count.readonly.post', True)
        self.assertJE(response, 'fields.books_count.readonly.put', True)
        self.assertJE(response, 'fields.books_count_annotate.type', None)
        self.assertJE(
            response, 'fields.books_count_annotate.name',
            'books_count_annotate')
        self.assertJE(
            response, 'fields.books_count_annotate.readonly.post', True)
        self.assertJE(
            response, 'fields.books_count_annotate.readonly.put', True)
        self.assertJE(response, 'fields.books_count_prop.type', None)
        self.assertJE(
            response, 'fields.books_count_prop.name', 'books_count_prop')
        self.assertJE(response, 'fields.books_count_prop.readonly.post', True)
        self.assertJE(response, 'fields.books_count_prop.readonly.put', True)

    def test_annotate_detail_service(self):
        response = self.client.get(f'/api/authors-annotate/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'books_count', 4)
        self.assertJE(response, 'books_count_annotate', 4)
        self.assertJE(response, 'books_count_prop', 4)

    def test_annotate_list_service(self):
        response = self.client.get(
            '/api/authors-annotate?items=1&count=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)
        self.assertJE(response, 'items.@0.last_name', 'Asimov')
        self.assertJE(response, 'items.@0.books_count', 3)
        self.assertJE(response, 'items.@0.books_count_annotate', 3)
        self.assertJE(response, 'items.@0.books_count_prop', 3)
        self.assertJE(response, 'items.@1.last_name', 'Dick')
        self.assertJE(response, 'items.@1.books_count', 0)
        self.assertJE(response, 'items.@1.books_count_annotate', 0)
        self.assertJE(response, 'items.@1.books_count_prop', 0)
        self.assertJE(response, 'items.@2.last_name', 'Rowling')
        self.assertJE(response, 'items.@2.books_count', 0)
        self.assertJE(response, 'items.@2.books_count_annotate', 0)
        self.assertJE(response, 'items.@2.books_count_prop', 0)
        self.assertJE(response, 'items.@3.last_name', 'Tolkien')
        self.assertJE(response, 'items.@3.books_count', 4)
        self.assertJE(response, 'items.@3.books_count_annotate', 4)
        self.assertJE(response, 'items.@3.books_count_prop', 4)

    def test_annotate_create_service(self):
        response = self.client.post(
            '/api/authors-annotate', {
                'first_name': 'Arthur Charles',
                'last_name': 'Clarke',
                'slug': 'arthur-c-clarke'})
        self.assertRC(response, 200)
        self.assertJE(response, 'books_count', 0)
        self.assertJE(response, 'books_count_annotate', 0)
        self.assertJE(response, 'books_count_prop', 0)

    def test_annotate_update_service(self):
        response = self.client.put(
            f'/api/authors-annotate/{self.author1.pk}', dict(
                first_name='J. R. R.',
                books_count=999,
                books_count_annotate=999,
                books_count_prop=999))

        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', 'J. R. R.')
        self.assertJE(response, 'books_count', 4)
        self.assertJE(response, 'books_count_annotate', 4)
        self.assertJE(response, 'books_count_prop', 4)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.first_name, 'J. R. R.')

    def test_fields_props_meta_service(self):
        response = self.client.get('/api/authors-fields-props/meta')
        self.assertRC(response, 200)
        self.assertJE(response, 'fields.first_name.type', 'CharField')
        self.assertJE(response, 'fields.first_name.name', 'First Name')
        self.assertJE(response, 'fields.books_count.type', 'IntegerField')
        self.assertJE(response, 'fields.books_count.name', 'Books Count')
        self.assertJE(response, 'fields.books_count.readonly.post', True)
        self.assertJE(response, 'fields.books_count.readonly.put', True)
        self.assertJE(
            response, 'fields.books_count_annotate.type', 'IntegerField')
        self.assertJE(
            response, 'fields.books_count_annotate.name',
            'Books Count (annotate)')
        self.assertJE(
            response, 'fields.books_count_annotate.readonly.post', True)
        self.assertJE(
            response, 'fields.books_count_annotate.readonly.put', True)
        self.assertJE(
            response, 'fields.books_count_prop.type', 'IntegerField')
        self.assertJE(
            response, 'fields.books_count_prop.name', 'Books Count (property)')
        self.assertJE(response, 'fields.books_count_prop.readonly.post', True)
        self.assertJE(response, 'fields.books_count_prop.readonly.put', True)

        response = self.client.get(
            '/api/authors-fields-props/meta?fields=true')
        self.assertRC(response, 200)
        self.assertJEExists(response, 'fields')
        response = self.client.get(
            '/api/authors-fields-props/meta?fields=false')
        self.assertRC(response, 200)
        self.assertJENotExists(response, 'fields')
        response = self.client.get(
            '/api/authors-fields-props/meta?other=true')
        self.assertRC(response, 200)
        self.assertJENotExists(response, 'fields')

    def test_fields_props_detail_service(self):
        response = self.client.get(
            f'/api/authors-fields-props/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'books_count', 4)
        self.assertJE(response, 'books_count_annotate', 4)
        self.assertJE(response, 'books_count_prop', 4)

    def test_fields_props_list_service(self):
        response = self.client.get(
            '/api/authors-fields-props?items=1&count=1&order=pk')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)
        self.assertJE(response, 'items.@0.books_count', 4)
        self.assertJE(response, 'items.@0.books_count_annotate', 4)
        self.assertJE(response, 'items.@0.books_count_prop', 4)
        self.assertJE(response, 'items.@1.books_count', 0)
        self.assertJE(response, 'items.@1.books_count_annotate', 0)
        self.assertJE(response, 'items.@1.books_count_prop', 0)

    def test_fields_props_create_service(self):
        response = self.client.post(
            '/api/authors-fields-props', {
                'first_name': 'Arthur Charles',
                'last_name': 'Clarke',
                'slug': 'arthur-c-clarke'})
        self.assertRC(response, 200)
        self.assertJE(response, 'books_count', 0)
        self.assertJE(response, 'books_count_annotate', 0)
        self.assertJE(response, 'books_count_prop', 0)

    def test_fields_props_update_service(self):
        response = self.client.put(
            f'/api/authors-fields-props/{self.author1.pk}', dict(
                first_name='J. R. R.',
                books_count=999,
                books_count_annotate=999,
                books_count_prop=999))

        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', 'J. R. R.')
        self.assertJE(response, 'books_count', 4)
        self.assertJE(response, 'books_count_annotate', 4)
        self.assertJE(response, 'books_count_prop', 4)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.first_name, 'J. R. R.')

    def test_alias_field(self):
        response = self.client.get(
            f'/api/books-custom-author/{self.author3_book1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'book_name', "The Caves of Steel")
        self.assertJE(response, 'author_last_name', "Asimov")

        response = self.client.get(
            f'/api/books-custom-author?author={self.author3.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'items.@0.author_last_name', "Asimov")
        self.assertJEExists(response, 'items.@0.book_name')

        response = self.client.post(
            '/api/books-custom-author', dict(
                author=self.author3.pk,
                book_name="A New Book",
                pub_date='2020-01-01'
            ))
        new_book_pk = response.json_content['pk']
        self.assertRC(response, 200)
        self.assertJE(response, 'book_name', "A New Book")
        self.assertJE(response, 'author_last_name', "Asimov")

        response = self.client.put(
            f'/api/books-custom-author/{new_book_pk}', dict(
                book_name="A New Book UPDATED"))
        self.assertRC(response, 200)
        self.assertJE(response, 'book_name', "A New Book UPDATED")
        self.assertJE(response, 'author_last_name', "Asimov")



import logging

from django.test import TestCase

from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests.models import Author, User

logger = logging.getLogger(__name__)


class PermsAPITest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient()
        response = self.client.post(
            '/api/auth/login', dict(
                username='jrr.tolkien',
                password='PASSWORD'))
        token = response.json_content['token']
        self.anonymous_config = dict(
            content_type='application/json')
        self.user1_config = dict(
            HTTP_AUTHORIZATION=f'Bearer {token}',
            content_type='application/json')

    @classmethod
    def setUpTestData(cls):
        cls.user1 = User.objects.create_user(
            username='jrr.tolkien',
            email="jrr.tolkien@oxford.com",
            password='PASSWORD',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
        )
        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien'
        )
        cls.author2 = Author.objects.create(
            first_name='Philip Kindred',
            last_name='Dick',
            slug='philip-k-dick'
        )
        cls.author3 = Author.objects.create(
            first_name='Isaac',
            last_name='Asimov',
            slug='isaac-asimov'
        )

    def test_private_edit_api_no_user(self):
        response = self.client.get(
            '/api/private-edit/authors', **self.anonymous_config)
        self.assertRC(response, 200)

        response = self.client.put(
            f'/api/private-edit/authors/{self.author1.pk}',
            {'first_name': 'J. R. R.'}, **self.anonymous_config)
        self.assertRC(response, 401)

    def test_private_edit_api_user(self):
        response = self.client.get(
            '/api/private-edit/authors', **self.user1_config)
        self.assertRC(response, 200)

        response = self.client.put(
            f'/api/private-edit/authors/{self.author1.pk}',
            {'first_name': 'J. R. R.'}, **self.user1_config)
        self.assertRC(response, 200)

    def test_private_api_no_user(self):
        response = self.client.get('/api/private/authors')
        self.assertRC(response, 401)

    def test_private_api_user(self):
        response = self.client.get(
            '/api/private/authors', **self.user1_config)
        self.assertRC(response, 200)

    def test_perm_api_no_admin(self):
        response = self.client.get(
            '/api/admin/authors', **self.user1_config)
        self.assertRC(response, 403)

    def test_perm_api_admin(self):
        self.user1.is_superuser = True
        self.user1.save()

        response = self.client.get(
            '/api/admin/authors', **self.user1_config)
        self.assertRC(response, 200)

    def test_perm_edit_api_no_admin(self):
        response = self.client.get(
            '/api/admin-edit/authors', **self.user1_config)
        self.assertRC(response, 200)

        response = self.client.put(
            f'/api/admin-edit/authors/{self.author1.pk}',
            {'first_name': 'J. R. R.'}, **self.user1_config)
        self.assertRC(response, 403)

    def test_perm_edit_api_admin(self):
        self.user1.is_superuser = True
        self.user1.save()

        response = self.client.get(
            '/api/admin-edit/authors', **self.user1_config)
        self.assertRC(response, 200)

        response = self.client.put(
            f'/api/admin-edit/authors/{self.author1.pk}',
            {'first_name': 'J. R. R.'}, **self.user1_config)
        self.assertRC(response, 200)



from datetime import date

from django.test import TestCase

from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests.models import Author, Book, BookType


class TestCache(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        cls.heroic_fantasy = BookType.objects.create(
            name="Heroic Fantasy")
        cls.author1_book1 = Book.objects.create(
            author=cls.author1,
            name="The Fellowship of the Ring",
            pub_date=date(1954, 1, 1),
            type=cls.heroic_fantasy)
        cls.author1_book2 = Book.objects.create(
            author=cls.author1,
            name="The Two Towers",
            pub_date=date(1954, 1, 1),
            type=cls.heroic_fantasy)
        cls.author1_book3 = Book.objects.create(
            author=cls.author1,
            name="The Return of the King",
            pub_date=date(1955, 1, 1),
            type=cls.heroic_fantasy)

    def test_cache_author_updated(self):
        Author.CACHE.clear()

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', True)

        response = self.client.put(f'/api/authors/{self.author1.pk}', dict(
            last_name="Updated"))
        self.assertRC(response, 200)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)
        self.assertJE(response, 'last_name', "Updated")

    def test_cache_author_book_updated(self):
        Author.CACHE.clear()

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', True)

        response = self.client.put(f'/api/books/{self.author1_book1.pk}', dict(
            name="Book Updated"))
        self.assertRC(response, 200)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)

    def test_cache_author_book_type_updated(self):
        Author.CACHE.clear()

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', True)

        response = self.client.put(
            f'/api/book-types/{self.heroic_fantasy.pk}', dict(
                name="Type Updated", slug="updated"))
        self.assertRC(response, 200)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)

    def test_cache_author_book_type_deleted(self):
        Author.CACHE.clear()

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', True)

        response = self.client.delete(
            f'/api/book-types/{self.heroic_fantasy.pk}')
        self.assertRC(response, 200)

        response = self.client.get(f'/api/authors/cache/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.from_cache', False)



from . import register_views, views

urlpatterns = register_views(
    views.LocaleRestView,
    views.AuthenticationView,
    views.SignupView,
    views.ForgottenPasswordView,
    views.OtpEmailView)



import inspect
import logging
from pathlib import Path

from django import apps
from django.apps import AppConfig
from django.db.models.signals import post_migrate

import yaml

from pfx.pfxcore.shortcuts import permissions, settings

logger = logging.getLogger(__name__)


def update_groups_permissions(
        groups, base_path, group_cache, is_last, module, out=True):
    from django.contrib.auth.models import Group

    out and print(f"Import file {base_path}/groups.yaml", end=" ")
    todo = False

    def log(txt):
        if not out:
            return
        nonlocal todo
        if not todo:
            print()
            todo = True
        print("  * ", txt)

    db_groups = {g.name: g for g in Group.objects.all()}

    for name, perms in groups.items():
        group_cache.add(name)
        created = name not in db_groups
        group = Group.objects.create(name=name) if created else db_groups[name]
        if created:
            log(f"Create group {name}")
        current_perms = {
            f'{p.content_type.app_label}.{p.codename}'
            for p in group.permissions.filter(content_type__app_label=module)}
        if created or not settings.PFX_AUTH_GROUPS_CREATE_ONLY:
            to_add = set(perms) - current_perms
            to_remove = current_perms - set(perms)
            if to_add:
                log(f"Add permissions for group {name}: {', '.join(to_add)}")
                group.permissions.add(*permissions(*to_add))
            if to_remove:
                log(f"Remove permissions for group {name}: "
                    f"{', '.join(to_remove)}")
                group.permissions.remove(*permissions(*to_remove))

    if is_last and not settings.PFX_AUTH_GROUPS_CREATE_ONLY:
        names = db_groups.keys() - group_cache
        if names:
            log(f"Delete groups: {', '.join(names)}")
            Group.objects.filter(name__in=names).delete()

    if not todo:
        out and print("[nothing to do]")


def update_groups_permissions_action(sender, **kwargs):
    if not isinstance(sender, PfxAppConfig):
        return

    base_path = '/'.join(sender.__class__.__module__.split('.')[:-1])
    is_last = [
        a for a in apps.apps.get_app_configs()
        if isinstance(a, PfxAppConfig)][-1].name == sender.name
    module = Path(inspect.getfile(sender.__class__)).parent.name

    groups_file = Path(
        Path(inspect.getfile(sender.__class__)).parent, 'groups.yaml')
    if not groups_file.exists():
        return

    with groups_file.open() as file:
        groups = yaml.safe_load(file)

    update_groups_permissions(
        groups, base_path, sender._processed_groups, is_last, module)


class PfxAppConfig(AppConfig):
    _processed_groups = set()

    def ready(self):
        post_migrate.connect(update_groups_permissions_action, sender=self)
        return super().ready()


class PfxCoreConfig(AppConfig):
    name = 'pfx.pfxcore'
    default = True



import logging

from django.test import TestCase
from django.test.utils import override_settings

from pfx.pfxcore.test import APIClient, TestAssertMixin

logger = logging.getLogger(__name__)


class TimezoneMiddlewareTest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient()

    def test_django_default_tz(self):
        response = self.client.get('/api/timezone')
        self.assertJE(response, "tz", 'America/Chicago')

    @override_settings(TIME_ZONE="Asia/Tokyo")
    def test_settings_default_tz(self):
        response = self.client.get('/api/timezone')
        self.assertJE(response, "tz", 'Asia/Tokyo')

    @override_settings(TIME_ZONE="Asia/Tokyo")
    def test_header_tz(self):
        response = self.client.get(
            '/api/timezone',  HTTP_X_CUSTOM_TIMEZONE="Europe/Zurich")
        self.assertJE(response, "tz", 'Europe/Zurich')

    @override_settings(TIME_ZONE="Asia/Tokyo")
    def test_header_unknown_tz(self):
        response = self.client.get(
            '/api/timezone',  HTTP_X_CUSTOM_TIMEZONE="Europe/Unknown")
        self.assertJE(response, "tz", 'Asia/Tokyo')



import json
from hashlib import sha1

from apispec.utils import deepupdate


class Schema:
    def __init__(self, id, name, **kwargs):
        self._id = id
        self._name = name
        self._kwargs = kwargs

    def id(self):
        hash = sha1(json.dumps(self.to_schema()).encode()).hexdigest()[:10]
        return f"{self._id}.{hash}"

    def to_ref(self, schema=False):
        return {
            'description': self._name,
            'content': {'application/json': {'schema':
                        schema and self.to_schema() or self.id()}}}

    def to_schema(self):
        return dict(**self._kwargs)


class ModelSchema(Schema):
    def __init__(self, model, fields, properties=None, mode=None):
        self.model = model
        self.fields = fields
        self._properties = properties
        self._mode = mode

    def id(self):
        hash = sha1(json.dumps(self.to_schema()).encode()).hexdigest()[:10]
        return f"{self.model.__name__}.{hash}"

    def to_ref(self, schema=False):
        return {
            'description': str(self.model._meta.verbose_name),
            'content': {'application/json': {'schema':
                        schema and self.to_schema() or self.id()}}}

    def default_json_repr_schema(self):
        from pfx.pfxcore.views.fields import FieldType
        return dict(
            pk=dict(
                type=FieldType.to_apidoc(FieldType.from_model_field(
                    self.model._meta.pk.__class__)),
                readonly=True),
            resource_name=dict(type='string', readonly=True))

    def to_schema(self):
        if self._mode in ('create', 'update'):
            properties = {}
        else:
            properties = getattr(
                self.model, 'json_repr_schema',
                self.default_json_repr_schema)()
        for name, field in self.fields.items():
            if self._mode == 'create' and field.readonly_create:
                continue
            if self._mode == 'update' and field.readonly_update:
                continue
            doc = (
                hasattr(self.model, 'apidoc') and
                self.model.apidoc.get(field.name) or {})
            properties[name] = deepupdate(
                field.to_apidoc(
                    self._mode in ('create', 'update')), doc.copy())
        if self._properties:
            properties.update(self._properties)
        return dict(properties=properties)


class ModelListSchema(ModelSchema):
    def to_ref(self, schema=False):
        return {
            'description': str(self.model._meta.verbose_name_plural),
            'content': {'application/json': {'schema':
                        schema and self.to_schema() or self.id()}}}

    def to_schema(self):
        model = super().to_schema()
        return dict(properties=dict(
            items=dict(type='array', items=dict(
                type='object', format=str(self.model._meta.verbose_name),
                **model)),
            meta=dict(type='object')))



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class ListItems(QueryParameter):
    """List of objects in response (`items`).

    * if `true` (default value), `items` is returned
    * if `false`, `items` is not returned

    If `count` attribute is set, the default value is `false`.
    """
    name = 'items'
    schema = dict(type='boolean', default=True)


register_global_parameter(ListItems)



from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.db import connection, models
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import include, path

from pfx.pfxcore import register_views
from pfx.pfxcore.decorator import rest_view
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.test import APIClient, TestAssertMixin
from pfx.pfxcore.views import VF, RestView
from tests.views import FakeViewMixin


class TestOne2ManyModel(JSONReprMixin, models.Model):
    name = models.CharField(max_length=30)

    class Meta:
        verbose_name = "TestOne2ManyModel"
        verbose_name_plural = "TestOne2ManyModel"
        ordering = ['pk']

    def clean_fields(self, exclude=None):
        errors = {}

        if 'rels' not in exclude and 'rels' in self._rel_data:
            if len(self._rel_data['rels']) < 2:
                errors.setdefault(NON_FIELD_ERRORS, []).append(
                    "You have to set minimum 2 rels.")

        try:
            super().clean_fields(exclude)
        except ValidationError as e:
            for k, ms in e.error_dict.items():
                errors.setdefault(k, []).extend(ms)

        if errors:
            raise ValidationError(errors)


class TestOne2ManyRelModel(JSONReprMixin, models.Model):
    name = models.CharField(max_length=30)
    descr = models.TextField()
    rel = models.ForeignKey(
        TestOne2ManyModel, related_name='rels', on_delete=models.CASCADE)

    class Meta:
        verbose_name = "TestOne2ManyRelModel"
        verbose_name_plural = "TestOne2ManyRelModels"
        ordering = ['pk']


@rest_view("/test")
class TestOne2ManyRestView(FakeViewMixin, RestView):
    default_public = True
    model = TestOne2ManyModel

    fields = ['name', VF('rels', related_fields=[
        'name', VF('descr', readonly=True)])]


urlpatterns = [
    path('api/', include(register_views(TestOne2ManyRestView))),
    path('api/', include('pfx.pfxcore.urls'))
]


@override_settings(ROOT_URLCONF=__name__)
class TestFieldsOne2Many(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(TestOne2ManyModel)
            schema_editor.create_model(TestOne2ManyRelModel)

    def test_meta(self):
        response = self.client.get('/api/test/meta')
        self.assertRC(response, 200)
        fields = 'fields.rels.fields'
        self.assertJE(response, f'{fields}.name.name', "name")
        self.assertJE(response, f'{fields}.name.type', "CharField")
        self.assertJE(response, f'{fields}.name.required', True)
        self.assertJE(response, f'{fields}.name.readonly.post', False)
        self.assertJE(response, f'{fields}.name.readonly.put', False)
        self.assertJE(response, f'{fields}.descr.name', "descr")
        self.assertJE(response, f'{fields}.descr.type', "TextField")
        self.assertJE(response, f'{fields}.descr.required', True)
        self.assertJE(response, f'{fields}.descr.readonly.post', True)
        self.assertJE(response, f'{fields}.descr.readonly.put', True)

    def test_get(self):
        o = TestOne2ManyModel.objects.create(name="Test")
        r1 = TestOne2ManyRelModel.objects.create(rel=o, name="R1", descr="D1")
        r2 = TestOne2ManyRelModel.objects.create(rel=o, name="R2", descr="D2")

        response = self.client.get(f'/api/test/{o.pk}')
        self.assertRC(response, 200)
        self.assertSize(response, 'rels', 2)
        self.assertJE(response, 'rels.@0.pk', r1.pk)
        self.assertJE(response, 'rels.@0.name', "R1")
        self.assertJE(response, 'rels.@0.descr', "D1")
        self.assertJE(response, 'rels.@1.pk', r2.pk)
        self.assertJE(response, 'rels.@1.name', "R2")
        self.assertJE(response, 'rels.@1.descr', "D2")

    def test_post(self):
        response = self.client.post('/api/test', dict(
            name="Test",
            rels=[dict(name="R1", descr="D1"), dict(name="R2", descr="D2")]
        ))
        self.assertRC(response, 200)
        self.assertSize(response, 'rels', 2)
        self.assertJE(response, 'rels.@0.name', "R1")
        self.assertJE(response, 'rels.@0.descr', "D1")
        self.assertJE(response, 'rels.@1.name', "R2")
        self.assertJE(response, 'rels.@1.descr', "D2")

    def test_post_field_errors(self):
        response = self.client.post('/api/test', dict(
            rels=[dict(name="R1", descr="D1"), dict(descr="D2")]
        ))
        self.assertRC(response, 422)
        self.assertEqual(
            set(response.json().keys()), {'name', 'rels::1::name'})
        self.assertJE(response, 'name', ["This field cannot be blank."])
        self.assertJE(
            response, 'rels::1::name', ["This field cannot be null."])

        response = self.client.post('/api/test', dict(
            rels=[dict(descr="D1")]
        ))
        self.assertRC(response, 422)
        self.assertEqual(
            set(response.json().keys()), {'__all__', 'name', 'rels::0::name'})
        self.assertJE(response, '__all__', ["You have to set minimum 2 rels."])
        self.assertJE(response, 'name', ["This field cannot be blank."])
        self.assertJE(
            response, 'rels::0::name', ["This field cannot be null."])

    def test_put(self):
        o = TestOne2ManyModel.objects.create(name="Test")
        r1 = TestOne2ManyRelModel.objects.create(rel=o, name="R1", descr="D1")

        vals = self.client.get(f'/api/test/{o.pk}').json()
        vals['rels'][0]['name'] = "R1 updated"
        vals['rels'].append({'name': "R2", "descr": "D2"})

        response = self.client.put(f'/api/test/{o.pk}', vals)
        self.assertRC(response, 200)
        self.assertSize(response, 'rels', 2)
        self.assertJE(response, 'rels.@0.pk', r1.pk)
        self.assertJE(response, 'rels.@0.name', "R1 updated")
        self.assertJE(response, 'rels.@0.descr', "D1")
        self.assertJE(response, 'rels.@1.name', "R2")
        self.assertJE(response, 'rels.@1.descr', "D2")

    def test_put_field_errors(self):
        o = TestOne2ManyModel.objects.create(name="Test")
        TestOne2ManyRelModel.objects.create(rel=o, name="R1", descr="D1")
        TestOne2ManyRelModel.objects.create(rel=o, name="R2", descr="D2")

        vals = self.client.get(f'/api/test/{o.pk}').json()
        vals['name'] = ""
        vals['rels'][0]['name'] = None
        del vals['rels'][1]

        response = self.client.put(f'/api/test/{o.pk}', vals)
        self.assertRC(response, 422)
        self.assertEqual(
            set(response.json().keys()), {'__all__', 'name', 'rels::0::name'})
        self.assertJE(response, '__all__', ["You have to set minimum 2 rels."])
        self.assertJE(response, 'name', ["This field cannot be blank."])
        self.assertJE(
            response, 'rels::0::name', ["This field cannot be null."])

    def test_put_delete(self):
        o = TestOne2ManyModel.objects.create(name="Test")
        r1 = TestOne2ManyRelModel.objects.create(rel=o, name="R1", descr="D1")
        r2 = TestOne2ManyRelModel.objects.create(rel=o, name="R2", descr="D2")
        TestOne2ManyRelModel.objects.create(rel=o, name="R3", descr="D3")

        vals = self.client.get(f'/api/test/{o.pk}').json()
        del vals['rels'][2]

        response = self.client.put(f'/api/test/{o.pk}', vals)
        self.assertRC(response, 200)
        self.assertSize(response, 'rels', 2)
        self.assertJE(response, 'rels.@0.pk', r1.pk)
        self.assertJE(response, 'rels.@0.name', "R1")
        self.assertJE(response, 'rels.@0.descr', "D1")
        self.assertJE(response, 'rels.@1.pk', r2.pk)
        self.assertJE(response, 'rels.@1.name', "R2")
        self.assertJE(response, 'rels.@1.descr', "D2")



PFX_TOKEN_SHORT_VALIDITY = {'hours': 12}
PFX_TOKEN_LONG_VALIDITY = {'days': 30}
PFX_TOKEN_OTP_VALIDITY = {'minutes': 15}
PFX_HOTP_CODE_VALIDITY = 15
PFX_OTP_VALID_WINDOW = 1
PFX_OTP_IMAGE = None
PFX_OTP_COLOR = None

PFX_COOKIE_DOMAIN = None
PFX_COOKIE_SECURE = True
PFX_COOKIE_SAMESITE = 'None'

PFX_LOGIN_BAN_FAILED_NUMBER = 5
PFX_LOGIN_BAN_SECONDS_START = 60
PFX_LOGIN_BAN_SECONDS_STEP = 60

PFX_OPENAPI_PATH = "doc/api/"
PFX_OPENAPI_FILENAME = "openapi"
PFX_OPENAPI_TEMPLATE = {}

PFX_PROFILING_SQL = ""
PFX_PROFILING_CPROFILE = ""

STORAGE_DEFAULT = None

STORAGE_S3_AWS_REGION = None
STORAGE_S3_AWS_ACCESS_KEY = None
STORAGE_S3_AWS_SECRET_KEY = None
STORAGE_S3_AWS_ENDPOINT_URL = None
STORAGE_S3_AWS_S3_BUCKET = None
STORAGE_S3_AWS_PUT_URL_EXPIRE = None
STORAGE_S3_AWS_GET_URL_EXPIRE = None

STORAGE_LOCAL_ROOT = None
STORAGE_LOCAL_X_ACCEL_REDIRECT = False

PFX_TEST_MODE = False

PFX_AUTH_GROUPS_CREATE_ONLY = False
